"""Module defining configuration classes for creating MetaFrames."""
from __future__ import annotations

from pathlib import Path
from typing import Iterable, Optional, Union

try:
    import tomllib
except ImportError:
    import tomli as tomllib  # type: ignore  # noqa

from metasyn.privacy import BasePrivacy, BasicPrivacy, get_privacy
from metasyn.provider import DistributionProviderList
from metasyn.varspec import VarSpec


class MetaConfig():
    """Configuration class for creating MetaFrames.

    This class is used to create, manipulate, and retrieve configurations for
    individual variables in a MetaFrame. It also provides methods for loading
    configurations from .toml files and converting them to dictionaries.

    Parameters
    ----------
    var_specs:
        List of configurations for individual variables. The order does not
        matter for variables that are found in the DataFrame, but in the case
        of variables that are data-free, the order is also the order of columns
        for the eventual synthesized dataframe. See the VarSpecAccess class on
        how the dictionary can be constructed.
    dist_providers:
        Distribution providers to use when fitting distributions to variables.
        Can be a string, provider, or provider type.
    privacy:
        Privacy method/level to use as a default setting for the privacy. Can be
        overridden in the var_spec for a particular column.
    n_rows:
        Number of rows for synthesization at a later stage. Can be unspecified by
        leaving the value at None.
    """

    def __init__(
            self,
            var_specs: Union[list[dict], list[VarSpec]],
            dist_providers: Union[DistributionProviderList, list[str], str, None],
            privacy: Optional[Union[BasePrivacy, dict]],
            n_rows: Optional[int] = None):
        self.var_specs = [self._parse_var_spec(v) for v in var_specs]
        self.dist_providers = dist_providers  # type: ignore
        self.privacy = privacy  # type: ignore
        self.n_rows = n_rows

    @staticmethod
    def _parse_var_spec(var_spec):
        if isinstance(var_spec, VarSpec):
            return var_spec
        return VarSpec.from_dict(var_spec)

    @property
    def privacy(self) -> BasePrivacy:
        """Return default privacy for generating the metaframe."""
        return self._privacy

    @privacy.setter
    def privacy(self, privacy: Optional[Union[dict, BasePrivacy]]):
        if privacy is None:
            self._privacy: BasePrivacy = BasicPrivacy()
        elif isinstance(privacy, dict):
            self._privacy = get_privacy(**privacy)
        else:
            self._privacy = privacy

    @property
    def dist_providers(self) -> DistributionProviderList:
        """Return the distribution provider list to be used for the metaframe."""
        return self._dist_providers

    @dist_providers.setter
    def dist_providers(self, dist_providers):
        if not isinstance(dist_providers, DistributionProviderList):
            self._dist_providers = DistributionProviderList(dist_providers)
        else:
            self._dist_providers = dist_providers

    @classmethod
    def from_toml(cls, config_fp: Union[str, Path]) -> MetaConfig:
        """Create a MetaConfig class from a .toml file.

        Parameters
        ----------
        config_fp:
            Path to the file containing the configuration.

        Returns
        -------
        meta_config:
            A fully initialized MetaConfig instance.
        """
        try:
            with open(config_fp, "rb") as handle:
                config_dict = tomllib.load(handle)
        except FileNotFoundError as fnf_error:
            raise FileNotFoundError(f"It appears '{config_fp}' is not a valid filepath."
                                    f" Please provide a path to a .toml file to load a MetaConfig"
                                    f" from.") from fnf_error
        except ValueError as value_error:
            if Path(config_fp).suffix == ".toml":
                raise ValueError(f"It appears '{Path(config_fp).name}' is a"
                                 f" '{Path(config_fp).suffix}' file."
                                 f" To load a MetaConfig, "
                                 f"provide it as a .toml file.") from value_error
            raise ValueError(f"An error occured while parsing the configuration file \n"
                             f"('{Path(config_fp).name}').") from value_error
        var_list = config_dict.pop("var", [])
        n_rows = config_dict.pop("n_rows", None)
        dist_providers = config_dict.pop("dist_providers", ["builtin"])
        privacy = config_dict.pop("privacy", {"name": "none", "parameters": {}})
        if len(config_dict) > 0:
            raise ValueError(f"Error parsing configuration file '{config_fp}'."
                             f" Unknown keys detected: '{list(config_dict)}'")
        return cls(var_list, dist_providers, privacy, n_rows=n_rows)

    def to_dict(self) -> dict:
        """Convert the configuration to a dictionary.

        Returns
        -------
        config_dict:
            Configuration in dictionary form.
        """
        return {
            "general": {
                "privacy": self.privacy,
                "dist_providers": self.dist_providers,
            },
            "var": self.var_specs
        }

    def get(self, name: str) -> VarSpecAccess:
        """Create a VarSpecAccess object pointing to a var with that name.

        If the variable does not exist, then a new variable config is created that
        has the default values.

        Parameters
        ----------
        name:
            Name of the variable configuration to retrieve.

        Returns
        -------
        var_spec:
            A variable config access object.
        """
        for var_spec in self.var_specs:
            if var_spec.name == name:
                return VarSpecAccess(var_spec, self)
        return VarSpecAccess(VarSpec(name=name), self)

    def iter_var(self, exclude: Optional[list[str]] = None) -> Iterable[VarSpecAccess]:
        """Iterate over all variables in the configuration.

        Parameters
        ----------
        exclude:
            Exclude variables with names in that list.

        Returns
        -------
        var_spec:
            VarSpecAccess class for each of the available variable configurations.
        """
        exclude = exclude if exclude is not None else []
        for var_spec in self.var_specs:
            if var_spec.name not in exclude:
                yield VarSpecAccess(var_spec, self)


class VarSpecAccess():  # pylint: disable=too-few-public-methods
    """Access for variable configuration object.

    They take into account what the defaults are from the MetaConfig object.
    Otherwise they pass through all the attributes as normal and thus behave
    exactly as a variable config object themselves.

    Parameters
    ----------
    var_spec
        The variable configuration to access.
    meta_config
        The meta configuration instance to get default values from.
    """

    def __init__(self, var_spec: VarSpec, meta_config: MetaConfig):
        self.var_spec = var_spec
        self.meta_config = meta_config

    def __getattribute__(self, attr):
        if attr == "privacy":
            if self.var_spec.privacy is None:
                return self.meta_config.privacy
            return self.var_spec.privacy
        if attr not in ("var_spec", "meta_config") and hasattr(self.var_spec, attr):
            return getattr(self.var_spec, attr)
        return super().__getattribute__(attr)



from pathlib import Path

import pytest
from pytest import mark

from metasyn.config import MetaConfig, VarSpecAccess
from metasyn.distribution import UniformDistribution
from metasyn.privacy import BasePrivacy
from metasyn.varspec import DistributionSpec, VarSpec


@mark.parametrize(
    "input,error",
    [
        ("uniform", False),
        (None, False),
        ({"implements": {"uniform": False}}, False),
        (UniformDistribution, False),
        (UniformDistribution(0, 2), False),
        (DistributionSpec(), False),
        ({"fit_kwargs": {"param": 3}}, True),
        ({"version": "2.0"}, True),
        ({"parameters": {"param": 2}}, True),
        (1, True),
    ]
)
def test_dist_spec(input, error):
    if error:
        with pytest.raises(Exception):
            DistributionSpec.parse(input)
    else:
        dist_spec = DistributionSpec.parse(input)
        assert isinstance(dist_spec, DistributionSpec)

def test_var_spec():
    var_spec = VarSpec("test", privacy={"name": "none", "parameters": {}})
    assert var_spec.name == "test"
    assert isinstance(var_spec.privacy, BasePrivacy)
    with pytest.raises(ValueError):
        var_spec = VarSpec("test", data_free=True)
    var_spec = VarSpec("test")
    assert isinstance(var_spec.dist_spec, DistributionSpec)
    var_spec = VarSpec.from_dict({"name": "test"})
    assert var_spec.name == "test"
    var_spec = VarSpec.from_dict({"name": "test", "distribution": "uniform"})
    assert var_spec.dist_spec.implements == "uniform"


def test_meta_config_datafree():
    meta_config = MetaConfig.from_toml(Path("tests", "data", "no_data_config.toml"))
    assert meta_config.n_rows == 100
    assert len(meta_config.var_specs) == 3
    assert isinstance(meta_config.var_specs[0], VarSpec)
    assert meta_config.var_specs[0].privacy is None
    assert isinstance(meta_config.var_specs[0].dist_spec, DistributionSpec)
    assert isinstance(meta_config.to_dict(), dict)
    var_spec = meta_config.get("PassengerId")
    assert isinstance(var_spec, VarSpecAccess)

    assert var_spec.data_free is True
    var_spec = meta_config.get("unknown")
    assert var_spec.name == "unknown"

    all_var_spec = list(meta_config.iter_var())
    assert len(all_var_spec) == 3
    assert isinstance(all_var_spec[0], VarSpecAccess)
    assert all_var_spec[0].meta_config == meta_config
    assert len(list(meta_config.iter_var(exclude=["PassengerId"]))) == 2


def test_meta_config():
    meta_config = MetaConfig.from_toml(Path("tests", "data", "example_config.toml"))
    assert len(meta_config.var_specs) == 5
    var_spec = meta_config.get("Cabin")
    assert var_spec.data_free is False
    assert var_spec.var_type is None
    assert var_spec.dist_spec.implements == "core.regex"



from pathlib import Path

import polars as pl
from pytest import mark, raises

from metasyn.demo.dataset import demo_file


@mark.parametrize("dataset", ["titanic"])
def test_demo_files(dataset):
    data_fp = demo_file(dataset)
    assert data_fp.is_file()


def test_demo_non_exist():
    with raises(ValueError):
        demo_file("non-existing-dataset")


def check_titan(output_fp):
    df = pl.read_csv(output_fp)
    assert len(df.columns) == 13
    assert "Board time" in df.columns


# def test_create_titanic(tmpdir):
#     output_fp = Path(tmpdir / "titanic.csv")
#     new_output_fp = create_titanic_demo(output_fp)
#     assert new_output_fp == output_fp
#     check_titan(output_fp)
#     new_output_fp = create_titanic_demo(output_fp)
#     assert new_output_fp == output_fp
#     check_titan(output_fp)



import pytest

from metasyn.distribution import MultinoulliDistribution
from metasyn.distribution.base import metadist
from metasyn.privacy import BasicPrivacy, get_privacy


@metadist(privacy="test")
class OtherMultinoulli(MultinoulliDistribution):
    pass


def test_base_privacy():
    privacy = BasicPrivacy()
    priv_dict = privacy.to_dict()
    assert isinstance(priv_dict, dict)
    assert priv_dict["name"] == "none"
    assert privacy.name == "none"
    assert isinstance(priv_dict["parameters"], dict)
    assert len(priv_dict["parameters"]) == 0
    assert privacy.is_compatible(MultinoulliDistribution)
    assert privacy.is_compatible(MultinoulliDistribution(["1"], [1.0]))
    assert not privacy.is_compatible(OtherMultinoulli)


def test_import_error():
    with pytest.raises(ImportError):
        get_privacy("disclosure")
    with pytest.raises(ImportError):
        get_privacy("dics")



"""Module defining MetaFrames, used for creating MetaFrames from DataFrames."""  # pylint: disable=invalid-name

from __future__ import annotations

import json
import pathlib
from datetime import datetime
from importlib.metadata import version
from typing import Any, Dict, List, Optional, Sequence, Union

import numpy as np
import polars as pl
from tqdm import tqdm

from metasyn.config import MetaConfig
from metasyn.privacy import BasePrivacy
from metasyn.validation import validate_gmf_dict
from metasyn.var import MetaVar
from metasyn.varspec import VarSpec


class MetaFrame():
    """Container for statistical metadata describing a dataset.

    This class is used to fit a MetaFrame to a Polars DataFrame, serialize and
    export the MetaFrame to a file, read a MetaFrame from a file, and create
    a synthetic Polars DataFrame.

    A MetaFrame represents a metadata frame, which is a structure that holds
    statistical metadata about a dataset. The data contained in a MetaFrame
    follows the Generative Metadata Format (GMF).
    The metadata is contained in a collection of MetaVar objects,
    with each MetaVar representing a column (variable).

    A MetaFrame can easily be created using the
    ``fit_dataframe`` method, which takes a Polars DataFrame and fits a
    MetaFrame to it.

    Parameters
    ----------
    meta_vars:
        List of variables representing columns in a DataFrame.
    n_rows:
        Number of rows in the original DataFrame.
    privacy_package:
        Package that supplies the distributions.
    """

    def __init__(self, meta_vars: List[MetaVar],
                 n_rows: Optional[int] = None):
        self.meta_vars = meta_vars
        self.n_rows = n_rows

    @property
    def n_columns(self) -> int:
        """int: Number of columns of the original dataframe."""
        return len(self.meta_vars)

    @classmethod
    def fit_dataframe(  # pylint: disable=too-many-branches
            cls,
            df: Optional[pl.DataFrame],
            var_specs: Optional[Union[list[VarSpec], pathlib.Path, str, MetaConfig]] = None,
            dist_providers: Optional[list[str]] = None,
            privacy: Optional[Union[BasePrivacy, dict]] = None,
            n_rows: Optional[int] = None,
            progress_bar: bool = True):
        """Create a metasyn object from a polars (or pandas) dataframe.

        The Polars dataframe should be formatted already with the correct
        datatypes, such as pl.Categorical (or the pandas equivalent).

        Parameters
        ----------
        df:
            Polars dataframe with the correct column dtypes.
        var_specs:
            Specifications for each column/variable. These specifications are supplied as
            a list of VarSpec instances (one for each column). Alternatively, the
            specifications can be entered as a path to a .toml file. For more information
            on this approach, see the MetaConfig class or the examples in the documentation.
            By default var_specs is None, which will use the default settings for each column.
        dist_providers:
            Distribution providers to use when fitting distributions to variables.
            Can be a string, provider, or provider type. This will overwrite the defaults if they
            were specified in the varspecs (but not the specifications per column).
        privacy:
            Privacy level to use by default. This will overwrite the defaults if they were
            specified in the varspecs (but not the specifications per column).
        n_rows:
            Number of rows registered in the MetaFrame. If left at None, it will use the number
            of rows in the input dataframe.
        progress_bar:
            Whether to create a progress bar.

        Returns
        -------
        MetaFrame:
            Initialized metasyn metaframe.
        """
        # Parse the var_specs into a MetaConfig instance.
        if isinstance(var_specs, (pathlib.Path, str)):
            meta_config = MetaConfig.from_toml(var_specs)
        elif isinstance(var_specs, MetaConfig):
            meta_config = var_specs
        elif var_specs is None:
            meta_config = MetaConfig([], dist_providers, privacy)
        else:
            meta_config = MetaConfig(var_specs, dist_providers, privacy)
        if dist_providers is not None:
            meta_config.dist_providers = dist_providers  # type: ignore
        if privacy is not None:
            meta_config.privacy = privacy  # type: ignore

        if df is not None and not isinstance(df, pl.DataFrame):
            if isinstance(df, (str, pathlib.Path)):
                raise ValueError("Please provide a DataFrame as input, not a string or path.")
            df = pl.DataFrame(df)
        all_vars = []
        columns = df.columns if df is not None else []
        if df is not None:
            for col_name in tqdm(columns, disable=not progress_bar):
                var_spec = meta_config.get(col_name)
                var = MetaVar.fit(
                    df[col_name],
                    var_spec.dist_spec,
                    meta_config.dist_providers,
                    var_spec.privacy,
                    var_spec.prop_missing,
                    var_spec.description)
                all_vars.append(var)

        # Data free columns to be appended
        for var_spec in meta_config.iter_var(exclude=columns):
            if not var_spec.data_free:
                raise ValueError(
                    f"Column with name '{var_spec.name}' not found and not declared as "
                     "data_free.")
            distribution = meta_config.dist_providers.create(var_spec)
            var = MetaVar(
                var_spec.name,
                var_spec.var_type,
                distribution,
                description=var_spec.description,
                prop_missing=var_spec.prop_missing,
            )
            all_vars.append(var)
        if df is None:
            if meta_config.n_rows is None:
                raise ValueError("Please provide the number of rows in the configuration, "
                                 "or supply a DataFrame.")
            return cls(all_vars, meta_config.n_rows)
        n_rows = len(df) if n_rows is None else n_rows
        return cls(all_vars, n_rows)

    @classmethod
    def from_config(cls, meta_config: MetaConfig) -> MetaFrame:
        """Create a MetaFrame using a configuration, but without a DataFrame.

        Parameters
        ----------
        meta_config
            Configuration to be used for creating the new MetaFrame.

        Returns
        -------
            A created MetaFrame.
        """
        return cls.fit_dataframe(None, meta_config)

    def to_dict(self) -> Dict[str, Any]:
        """Create dictionary with the properties for recreation."""
        return {
            "n_rows": self.n_rows,
            "n_columns": self.n_columns,
            "provenance": {
                "created by": {
                    "name": "metasyn",
                    "version": version("metasyn"),
                },
                "creation time": datetime.now().isoformat()
            },
            "vars": [var.to_dict() for var in self.meta_vars],
        }

    def __getitem__(self, key: Union[int, str]) -> MetaVar:
        """Return meta var either by variable name or index."""
        if isinstance(key, int):
            # If the key is an integer, give the ith variable.
            return self.meta_vars[key]
        if isinstance(key, str):
            # If the key is a string, return the first variable with that name.
            for var in self.meta_vars:
                if var.name == key:
                    return var
            raise KeyError(f"Cannot find variable '{key}'")
        raise TypeError(f"Cannot get item for key '{key}'")

    def __str__(self) -> str:
        """Return an easy to read formatted string for the metaframe."""
        vars_formatted = "\n".join(
            f"Column {i + 1}: {str(var)}" for i,
            var in enumerate(
                self.meta_vars))
        return (
            f"# Rows: {self.n_rows}\n"
            f"# Columns: {self.n_columns}\n\n"
            f"{vars_formatted}\n"
        )

    @property
    def descriptions(self) -> dict[str, str]:
        """Return the descriptions of the columns."""
        return {var.name: var.description for var in self.meta_vars
                if var.name is not None and var.description is not None}

    @descriptions.setter
    def descriptions(
            self, new_descriptions: Union[dict[str, str], Sequence[str]]):
        if isinstance(new_descriptions, dict):
            for var_name, new_desc in new_descriptions.items():
                self[var_name].description = new_desc
        else:
            assert len(new_descriptions) == self.n_columns, (
                "Descriptions need to be either a dict or a "
                "sequence with the length of the number of variables.")
            for i_desc, new_desc in enumerate(new_descriptions):
                self[i_desc].description = new_desc

    def export(self, fp: Optional[Union[pathlib.Path, str]],
               validate: bool = True) -> None:
        """Serialize and export the MetaFrame to a JSON file, following the GMF format.

        Optionally, validate the exported JSON file against the JSON schema(s) included in the
        package.

        Parameters
        ----------
        fp:
            File to write the metaframe to.
        validate:
            Validate the JSON file with a schema.
        """
        self_dict = _jsonify(self.to_dict())
        if validate:
            validate_gmf_dict(self_dict)
        if fp is None:
            print(json.dumps(self_dict, indent=4))
        else:
            with open(fp, "w", encoding="utf-8") as f:
                json.dump(self_dict, f, indent=4)

    def to_json(self, fp: Union[pathlib.Path, str],
                validate: bool = True) -> None:
        """Serialize and export the MetaFrame to a JSON file, following the GMF format.

        This method is a wrapper and simply calls the 'export' function.

        Optionally, validate the exported JSON file against the JSON schema(s) included in the
        package.

        Parameters
        ----------
        fp:
            File to write the metaframe to.
        validate:
            Validate the JSON file with a schema.
        """
        self.export(fp, validate)

    @classmethod
    def from_json(cls, fp: Union[pathlib.Path, str],
                  validate: bool = True) -> MetaFrame:
        """Read a MetaFrame from a JSON file.

        Parameters
        ----------
        fp:
            Path to read the data from.
        validate:
            Validate the JSON file with a schema.

        Returns
        -------
        MetaFrame:
            A restored MetaFrame from the file.
        """
        with open(fp, "r", encoding="utf-8") as f:
            self_dict = json.load(f)

        if validate:
            validate_gmf_dict(self_dict)

        n_rows = self_dict["n_rows"]
        meta_vars = [MetaVar.from_dict(d) for d in self_dict["vars"]]
        return cls(meta_vars, n_rows)

    def synthesize(self, n: Optional[int] = None) -> pl.DataFrame:
        """Create a synthetic Polars dataframe.

        Parameters
        ----------
        n:
            Number of rows to generate, if None, use number of rows in original dataframe.

        Returns
        -------
        polars.DataFrame:
            Dataframe with the synthetic data.
        """
        if n is None:
            if self.n_rows is None:
                raise ValueError("Cannot synthesize DataFrame, since number of rows is unknown."
                                 "Please specify the number of rows to synthesize.")
            n = self.n_rows
        synth_dict = {var.name: var.draw_series(n) for var in self.meta_vars}
        return pl.DataFrame(synth_dict)

    def __repr__(self) -> str:
        """Return the MetaFrame as it would be output to JSON."""
        pretty_data = _jsonify(self.to_dict())
        output = json.dumps(pretty_data, indent=4)
        return output


def _jsonify(data):
    if isinstance(data, (list, tuple)):
        return [_jsonify(d) for d in data]
    if isinstance(data, dict):
        return {key: _jsonify(value) for key, value in data.items()}

    if isinstance(data, (np.int8, np.int16, np.int32, np.int64)):
        return int(data)
    if isinstance(data, np.ndarray):
        return _jsonify(data.tolist())
    return data



"""Module all string distributions."""
from __future__ import annotations

from typing import Iterable, Optional, Union

# from lingua._constant import LETTERS, PUNCTUATION
import regex
from faker import Faker
from lingua import LanguageDetectorBuilder  # pylint: disable=no-name-in-module
from regexmodel import NotFittedError, RegexModel
from scipy.stats import poisson

from metasyn.distribution.base import (
    BaseConstantDistribution,
    BaseDistribution,
    UniqueDistributionMixin,
    metadist,
)

LETTERS = regex.compile(r"\p{Han}|\p{Hangul}|\p{Hiragana}|\p{Katakana}|\p{L}+")
PUNCTUATION = regex.compile(r"\p{P}")

@metadist(implements="core.faker", var_type="string")
class FakerDistribution(BaseDistribution):
    """Faker distribution for cities, addresses, etc.

    This is mainly an interface for the faker package, so that it
    can be used within the metasyn package. It doesn't have any
    true fitting/statistical inference method, so it has to be manually
    selected.

    Parameters
    ----------
    faker_type: str
        The provider function in the faker package, e.g. 'city' or 'ipv4', etc.
    locale: str
        Locale used for the faker package.

    Examples
    --------
    >>> FakerDistribution(faker_type="city", locale="en_US")
    >>> FakerDistribution(faker_type="address", locale="nl_NL")
    """

    def __init__(self, faker_type: str, locale: str = "en_US"):
        self.faker_type: str = faker_type
        self.locale: str = locale
        self.fake: Faker = Faker(locale=locale)

    @classmethod
    def _fit(cls, values, faker_type: str = "city", locale: str = "en_US"):  \
            # pylint: disable=arguments-differ
        """Select the appropriate faker function and locale."""
        return cls(faker_type, locale)

    def draw(self):
        return getattr(self.fake, self.faker_type)()

    def information_criterion(self, values: Iterable) -> float:
        return 99999

    @classmethod
    def default_distribution(cls):
        return cls("city")

    def _param_dict(self):
        return {
            "faker_type": self.faker_type,
            "locale": self.locale
        }

    @classmethod
    def _param_schema(cls):
        return {
            "faker_type": {"type": "string"},
            "locale": {"type": "string"},
        }


@metadist(implements="core.faker", var_type="string")
class UniqueFakerDistribution(UniqueDistributionMixin, FakerDistribution):
    """Faker distribution that returns unique values.

    See :class:`~FakerDistribution` for examples and explanation.
    """


@metadist(implements="core.freetext", var_type="string")
class FreeTextDistribution(BaseDistribution):
    """Free text distribution.

    This distribution detects the language and generates sentences using
    the Faker package. The average number of sentences and words per item
    are detected using regexes (with the lingua package).

    Parameters
    ----------
    locale: str
        Locale used for the faker package.
    avg_sentences:
        Average number of sentences (punctuation marks) per (non-NA) row,
        if None do not make sentences.
    avg_words:
        Average number of words per (non-NA) row.
    """

    def __init__(self, locale: str, avg_sentences: Optional[float], avg_words: float):
        self.locale: str = locale
        self.avg_sentences = avg_sentences
        self.avg_words = avg_words
        self.fake = Faker(locale=self.locale)

    @classmethod
    def _fit(cls, values, max_values: int = 50):
        """Select the appropriate faker function and locale."""
        lang_str = cls.detect_language(values[:max_values])
        if lang_str is None:
            return cls.default_distribution()

        try:
            Faker(lang_str)
        except AttributeError:
            lang_str = "EN"

        all_text = "\n".join(values)
        n_non_empty = (values != "").sum()
        n_punctuation = len(list(PUNCTUATION.finditer(all_text)))
        n_words = len(list(LETTERS.finditer(all_text)))
        if n_punctuation < n_non_empty//3:
            avg_sentence = None
        else:
            avg_sentence = n_punctuation/len(values)
        avg_words = n_words/len(values)
        return cls(lang_str, avg_sentence, avg_words)

    @classmethod
    def detect_language(cls, values: Iterable) -> Optional[str]:
        """Detect the language of some text.

        Parameters
        ----------
        values:
            Values to detect the language of (usually polars dataframe).

        Returns
        -------
        language:
            Two letter ISO code to represent the language, or None if it could not be determined.
        """
        detector = LanguageDetectorBuilder.from_all_languages().with_low_accuracy_mode().build()
        lang = detector.detect_language_of("\n".join(values))
        if lang is None:
            return None
        return str(lang.iso_code_639_1).rsplit(".", maxsplit=1)[-1]

    def draw(self):
        if self.avg_sentences is None:
            n_words = max(1, poisson(self.avg_words).rvs())
            sentence = self.fake.sentence(n_words)
            return sentence[:-1]

        n_sentences = max(1, poisson(self.avg_sentences).rvs())
        avg_words_per_sent = max(1, self.avg_words/max(1, self.avg_sentences))
        n_words = max(1, poisson(avg_words_per_sent).rvs())
        return " ".join(self.fake.sentence(nb_words=n_words) for _ in range(n_sentences))

    def information_criterion(self, values) -> float:
        series = self._to_series(values)
        # Check the average number of characters
        avg_chars = series.str.len_chars().mean()
        if avg_chars is not None and avg_chars >= 25:  # type: ignore  # Workaround polars typing
            lang = self.detect_language(series)
            if lang is not None:
                return -1.0
        return 99999999

    @classmethod
    def default_distribution(cls):
        return cls("en_US", 3, 6)

    def _param_dict(self):
        return {
            "locale": self.locale,
            "avg_sentences": self.avg_sentences,
            "avg_words": self.avg_words,
        }

    @classmethod
    def _param_schema(cls):
        return {
            "locale": {"type": "string"},
            "avg_sentences": {"type": ["number", "null"]},
            "avg_words": {"type": "number"},
        }


@metadist(implements="core.constant", var_type="string")
class StringConstantDistribution(BaseConstantDistribution):
    """Constant string distribution.

    This class implements the constant distribution, so that it draws always
    the same value.

    Parameters
    ----------
    value: str
        Value that will be returned when drawn.

    Examples
    --------
    >>> ConstantDistribution("some_string")
    """

    @classmethod
    def default_distribution(cls) -> BaseDistribution:
        return cls("text")

    @classmethod
    def _param_schema(cls):
        return {
            "value": {"type": "string"}
        }

@metadist(implements="core.regex", var_type="string", version="2.0")
class RegexDistribution(BaseDistribution):
    """Structured string distribution using regex.

    Main implementation details in the regexmodel package:
    https://github.com/sodascience/regexmodel

    This distribution tries to create a regex that fits the variable. This
    regex also contains statistical information about the probabilities if the
    regex has multiple options (e.g. a|b). The regex is only a subset what is provided
    by the python re package. What is currently implemented:

    - Parentheses with multiple options and no modifiers, e.g. ([a]|[b]|[c])
    - Square brackets without negation, e.g. [abc]
    - Ranges [A-Z], [a-z], [0-9], but not subranges (e.g. [0-3])
    - Repetition quantifiers (curly brackets) with minimum and maximum [A-Z]{3,6}, but not [A-Z]{6}.

    When fitting the RegexDistribution using the fit method, pay attention to the
    count_thres and method parameters. By default these will be dynamic and take reasonable
    values for the input, but in some cases it can be important to set them manually.
    The count_thres parameter sets the minimum number of times a regex element needs to
    be used. So, if count_thres=2, and there is only one value starting with "a", then the regex
    will never start with "a". In effect, a higher value will provide more privacy, less
    utility and a faster fit. The other parameter "method" has a small effect on the accuracy
    of the regex, and a larger effect on the worst case time consumption for fitting.
    Set to "accurate" for the best result, and "fast" for the fastest result.

    Examples that this distribution should work reasonably for are: email,
    ID's, telephone numbers, ip addresses, etc.

    Parameters
    ----------
    regex_data:
        Valid inputs for the regex model are:
        - str: String with a regex (that falls within the specifications, see above).
        - dict: Serialized version of the regex model, as it is coming from a JSON file.
        - RegexModel: Initialized regex model.

    Examples
    --------
    >>> RegexDistribution(r"AB[0-9]{4}").draw()
    "AB8123"
    >>> RegexDistribution(r"(a|b|c)10)").draw()
    "b10"
    """

    def __init__(self, regex_data: Union[str, dict, RegexModel]):
        self.regex_model = RegexModel(regex_data)

    @classmethod
    def _fit(cls, values, count_thres: Optional[int] = None, method: str = "auto"):
        """Fit a regex to structured strings.

        Arguments
        ---------
        values:
            Values to be fitted (pl.Series).
        count_thres:
            Threshold for regex elements, so that a regex element can only be used if
            the number of values satisfying said element is higher than the threshold.
        method:
            Method for fitting the regex model. Possible values are ["accurate", "fast", "auto"]
            The "auto" method switches between the "accurate" and "fast" methods depending on
            the number of characters (fast if #char > 10000) in the series.
        """
        if method == "auto":
            if values.str.len_chars().mean() > 10:
                method = "fast"
            else:
                method = "accurate"

        # Make count_thres ~= #values/100 up to 50 if in auto mode.
        if count_thres is None:
            count_thres = min(50, max(2, round(len(values)/50)))

        # Try to fit the values, if it cannot be fit, then use the default distribution.
        try:
            model = RegexModel.fit(values, count_thres=count_thres, method=method)
        except NotFittedError:
            return cls.default_distribution()
        return cls(model)

    def draw(self):
        return self.regex_model.draw()

    def _param_dict(self):
        return {"regex_data": self.regex_model.serialize()}

    @property
    def _params_formatted(self):
        return f"\t- regex: {self.regex_model.regex}"

    @classmethod
    def _param_schema(cls):
        return {
            "regex": {
                "type": "string"
            },
            "counts": {
                "type": "array"
            }
        }

    @classmethod
    def default_distribution(cls):
        return cls(r"[ABC][0-9]{3,4}")


@metadist(implements="core.regex", var_type="string")
class UniqueRegexDistribution(UniqueDistributionMixin, RegexDistribution):
    """Unique variant of the regex distribution.

    See :class:`~RegexDistribution` for examples and explanation on this distribution.
    """



"""
Module serving as the basis for all metasyn distributions.

The base module contains the ``BaseDistribution`` class,
which is the base class for all distributions.
It also contains the ``ScipyDistribution`` class,
which is a specialized base class for distributions that are built on top of
SciPy's statistical distributions.

Additionally it contains the ``UniqueDistributionMixin`` class,
which is a mixin class that can be used to make a distribution unique
(i.e., one that does not contain duplicate values).

Finally it contains the ``metadist()`` decorator,
which is used to set the class attributes of a distribution.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Optional, Union

import numpy as np
import polars as pl
from numpy import Inf
from numpy import typing as npt


class BaseDistribution(ABC):
    """Abstract base class to define a distribution.

    All distributions should be derived from this class, and should implement
    the following methods:
    :meth:`~_fit`,
    :meth:`~draw`,
    :meth:`~_param_dict`,
    :meth:`~_param_schema`,
    :meth:`~default_distribution`
    and ``__init__``.
    """

    implements: str = "unknown"
    """The identifier for the implemented distribution"""
    var_type: str = "unknown"
    """The variable type of the distribution"""
    provenance: str = "builtin"
    """Which plugin provides the distribution or builtin"""
    privacy: str = "none"
    """The type of privacy it implements"""
    unique: bool = False
    """Whether the distribution creates only unique values"""
    version: str = "1.0"
    """Version of the implemented distribution"""

    @classmethod
    def fit(cls, series: Union[pl.Series, npt.NDArray],
            *args, **kwargs) -> BaseDistribution:
        """Fit the distribution to the series.

        Parameters
        ----------
        series: polars.Series
            Data to fit the distribution to.

        Returns
        -------
        BaseDistribution:
            Fitted distribution.
        """
        pl_series = cls._to_series(series)
        if len(pl_series) == 0:
            return cls.default_distribution()
        return cls._fit(pl_series, *args, **kwargs)

    @staticmethod
    def _to_series(values: Union[npt.NDArray, pl.Series]) -> pl.Series:
        if not isinstance(values, (pl.Series, np.ndarray)):
            values = pl.Series(values)
        if isinstance(values, pl.Series):
            series = values.drop_nulls()
        else:
            series_array = np.array(values)
            series_array = series_array[~np.isnan(series_array)]
            series = pl.Series(series_array)
        return series

    @classmethod
    @abstractmethod
    def _fit(cls, values: pl.Series) -> BaseDistribution:
        """See fit method, but does not need to deal with NA's."""

    @abstractmethod
    def draw(self) -> object:
        """Draw a random element from the fitted distribution."""

    def draw_reset(self) -> None:
        """Reset the drawing of elements to start again."""

    @property
    def _params_formatted(self) -> str:
        return "\n".join(
            f"\t- {param}: {value}" for param,
            value in self._param_dict().items()
        )

    def __str__(self) -> str:
        """Return an easy to read formatted string for the distribution."""
        return (
            f"- Type: {self.implements}\n"
            f"- Provenance: {self.provenance}\n"
            f"- Parameters:\n"
            f"{self._params_formatted}\n"
        )

    @abstractmethod
    def _param_dict(self):
        """Get dictionary with the parameters of the distribution."""

    def to_dict(self) -> dict:
        """Convert the distribution to a dictionary."""
        return {
            "implements": self.implements,
            "version": self.version,
            "provenance": self.provenance,
            "class_name": self.__class__.__name__,
            "unique": self.unique,
            "parameters": deepcopy(self._param_dict()),
        }

    @classmethod
    @abstractmethod
    def _param_schema(cls):
        """Get schema for the parameters of the distribution."""

    @classmethod
    def schema(cls) -> dict:
        """Create sub-schema to validate GMF file."""
        return {
            "type": "object",
            "properties": {
                "implements": {"const": cls.implements},
                "version": {"type": "string"},
                "provenance": {"const": cls.provenance},
                "class_name": {"const": cls.__name__},
                "unique": {"const": cls.unique},
                "parameters": {
                    "type": "object",
                    "properties": cls._param_schema(),
                    "required": list(cls.default_distribution()._param_dict())
                }
            },
            "required": ["implements", "provenance", "class_name", "parameters"]
        }

    @classmethod
    def from_dict(cls, dist_dict: dict) -> BaseDistribution:
        """Create a distribution from a dictionary."""
        return cls(**dist_dict["parameters"])

    def information_criterion(self, values: Union[pl.Series, npt.NDArray]) -> float:  # pylint: disable=unused-argument
        """Get the BIC value for a particular set of values.

        Parameters
        ----------
        values: array_like
            Values to determine the BIC value of.
        """
        return 0.0

    @classmethod
    def matches_name(cls, name: str) -> bool:
        """Check whether the name matches the distribution.

        Parameters
        ----------
        name: str
            Name to match to the distribution.

        Returns
        -------
        bool:
            Whether the name matches.
        """
        assert cls.implements != "unknown", f"Internal error in class {cls.__name__}"
        return name in (cls.implements.split(".")[1],
                        cls.implements,
                        cls.__name__,
                        )

    @classmethod
    @abstractmethod
    def default_distribution(cls) -> BaseDistribution:
        """Get a distribution with default parameters."""
        return cls()


def metadist(
        implements: Optional[str] = None,
        provenance: Optional[str] = None,
        var_type: Optional[Union[str, list[str]]] = None,
        unique: Optional[bool] = None,
        version: Optional[str] = None,
        privacy: Optional[str] = None):
    """Decorate class to create a distribution with the right properties.

    Parameters
    ----------
    implements:
        The distribution ID that it implements, e.g. core.uniform, core.regex.
    provenance:
        Where the distribution came from, which package/plugin implemented it.
    var_type:
        Variable type of the distribution, e.g. continuous, categorical, string.
    unique:
        Whether the distribution is unique or not.
    privacy:
        Privacy class/implementation of the distribution.
    version:
        Version of the distribution. Increment this to ensure that compatibility is
        properly handled.

    Returns
    -------
    cls:
        Class with the appropriate class variables.
    """
    def _wrap(cls):
        if implements is not None:
            cls.implements = implements
        if provenance is not None:
            cls.provenance = provenance
        if var_type is not None:
            cls.var_type = var_type
        if unique is not None:
            cls.unique = unique
        if privacy is not None:
            cls.privacy = privacy
        if version is not None:
            cls.version = version
        if cls.__doc__ is None:
            return cls
        cls.__doc__ = cls.__doc__.rstrip(" ")
        if not cls.__doc__.endswith("\n"):
            cls.__doc__ += "\n"
        cls.__doc__ += f"""
    Attributes
    ----------
    implements:
        {cls.implements}
    unique:
        {cls.unique}
    version:
        {cls.version}
    var_type:
        {cls.var_type}
    privacy:
        {cls.privacy}
    provenance:
        {cls.provenance}
    """
        return cls

    return _wrap


class ScipyDistribution(BaseDistribution):
    """Base class for numerical distributions using Scipy.

    This base class makes it easy to implement new numerical
    distributions. It can also be used for non-Scipy distributions,
    provided the distribution implements `logpdf`, `rvs` and `fit` methods.
    """

    @property
    def n_par(self) -> int:
        """int: Number of parameters for distribution."""
        return len(self.par)

    def __getattr__(self, attr: str):
        """Get attribute for easy access to parameters.

        Parameters
        ----------
        attr:
            Attribute to retrieve. If the attribute is a parameter
            name, then retrieve that parameter, otherwise use the default
            implementation for getting the attribute.

        Returns
        -------
        object:
            Parameter or attribute.
        """
        if attr != "par" and attr in self.par:
            return self.par[attr]
        return object.__getattribute__(self, attr)

    @classmethod
    def _fit(cls, values):
        if len(values) == 0:
            return cls.default_distribution()
        param = cls.dist_class.fit(values)  # type: ignore  # All derived classes have dist_class
        return cls(*param)

    def _param_dict(self):
        return self.par

    def draw(self):
        return self.dist.rvs()

    def information_criterion(self, values):
        vals = self._to_series(values)
        if len(vals) == 0:
            return 2 * self.n_par
        return self._information_criterion(vals)

    def _information_criterion(self, values):
        return np.log(len(values)) * self.n_par - 2 * np.sum(self.dist.logpdf(values))


@metadist(unique=True)
class UniqueDistributionMixin(BaseDistribution):
    """Mixin class to make unique version of base distributions.

    This mixin class can be used to extend base distribution classes, adding
    functionality that ensures generated values are unique. It overrides
    the `draw` method of the base class, adding a check to prevent duplicate
    values from being drawn. If a duplicate value is drawn, it retries up to
    1e5 times before raising a ValueError.

    The `UniqueDistributionMixin` is used in various unique metasyn distribution
    variations, such as `UniqueFakerDistribution` and `UniqueRegexDistribution`.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.key_set: set = set()

    def draw_reset(self):
        self.key_set = set()

    def draw(self) -> object:
        n_retry = 0
        while n_retry < 1e5:
            new_val = super().draw()  # type: ignore
            if new_val not in self.key_set:
                self.key_set.add(new_val)
                return new_val
            n_retry += 1
        raise ValueError(f"Failed to draw unique string after {n_retry} tries.")

    def information_criterion(self, values):
        return 9999999



class BaseConstantDistribution(BaseDistribution):
    """Base class for constant distribution.

    This base class makes it easy to implement new constant distributions
    for different variable types.
    """

    def __init__(self, value) -> None:
        self.value = value

    @property
    def n_par(self) -> int:
        """int: Number of parameters for distribution."""
        return 0

    @classmethod
    def _fit(cls, values: pl.Series) -> BaseDistribution:
        # if unique, just get that value
        if values.n_unique() == 1:
            return cls(values.unique()[0])

        # otherwise get most common value
        val = values.value_counts(sort=True)[0,0]
        return cls(val)

    def _param_dict(self):
        return {"value": self.value}

    def draw(self):
        return self.value

    def information_criterion(self, values):
        vals = self._to_series(values)
        return -Inf if vals.n_unique() < 2 else Inf



import pytest
from pytest import mark

from metasyn.distribution import MultinoulliDistribution, UniformDistribution
from metasyn.distribution.base import BaseDistribution, metadist
from metasyn.provider import BuiltinDistributionProvider, DistributionProviderList


@mark.parametrize("input", ["builtin", "fake-name", BuiltinDistributionProvider,
                            BuiltinDistributionProvider()])
def test_plist_init(input):
    if input == "fake-name":
        with pytest.raises(ValueError):
            DistributionProviderList(input)
    else:
        plist = DistributionProviderList(input)
        assert issubclass(plist.find_distribution("core.regex", var_type="string"), BaseDistribution)


@metadist(version="1.0")
class UniformTest1(UniformDistribution):
    pass


@metadist(version="1.1")
class UniformTest11(UniformDistribution):
    pass


@metadist(version="2.0")
class UniformTest2(UniformDistribution):
    pass


class CheckProvider(BuiltinDistributionProvider):
    distributions = [UniformTest2]
    legacy_distributions = [UniformTest1, UniformTest11]


class LegacyOnly(BuiltinDistributionProvider):
    distributions = [MultinoulliDistribution]
    legacy_distributions = [UniformTest1, UniformTest11, UniformTest2]


def test_legacy():
    plist = DistributionProviderList(CheckProvider)
    assert issubclass(plist.find_distribution("core.uniform", var_type="continuous"), UniformTest2)
    with pytest.warns():
        assert issubclass(plist.find_distribution("core.uniform", var_type="continuous", version="1.1"), UniformTest11)
    with pytest.warns():
        assert issubclass(plist.find_distribution("core.uniform", var_type="continuous", version="1.0"), UniformTest1)
    with pytest.warns():
        assert issubclass(plist.find_distribution("core.uniform", var_type="continuous", version="1.2"), UniformTest11)
    with pytest.raises(ValueError):
        plist.find_distribution("core.uniform", var_type="continuous", version="0.9")
    with pytest.warns():
        assert issubclass(plist.find_distribution("core.uniform", var_type="continuous", version="2.1"), UniformTest2)

    plist = DistributionProviderList(LegacyOnly)
    with pytest.warns():
        assert issubclass(plist.find_distribution("core.uniform", var_type="continuous"), UniformTest2)



from pytest import mark, raises

from metasyn.distribution import UniformDistribution
from metasyn.privacy import BasicPrivacy
from metasyn.provider import get_distribution_provider
from metasyn.testutils import check_distribution, check_distribution_provider


def test_builtin_provider():
    check_distribution_provider("builtin")


@mark.parametrize(
    "distribution", get_distribution_provider("builtin").distributions
)
def test_dist_validation(distribution):
    check_distribution(distribution, privacy=BasicPrivacy(),
                       provenance="builtin")


class Distribution(UniformDistribution):
    def schema():
        return "[{"


def test_schema_val():
    with raises(ValueError):
        check_distribution(Distribution, "none", "builtin")



import datetime as dt
import json

import numpy as np
import pandas as pd
import polars as pl
import pytest
from pytest import mark, raises

from metasyn.distribution import (
    DiscreteUniformDistribution,
    NormalDistribution,
    RegexDistribution,
    UniformDistribution,
    UniqueRegexDistribution,
)
from metasyn.distribution.categorical import MultinoulliDistribution
from metasyn.distribution.continuous import TruncatedNormalDistribution
from metasyn.distribution.discrete import UniqueKeyDistribution
from metasyn.metaframe import _jsonify
from metasyn.var import MetaVar


def _series_drop_nans(series):
    if isinstance(series, pl.Series):
        return series.drop_nulls()
    return series.dropna()


def _series_element_classname(series):
    series = _series_drop_nans(series)
    if isinstance(series, pl.Series):
        return series[0].__class__.__name__
    return series.iloc[0].__class__.__name__


def check_var(series, var_type, temp_path):
    def check_similar(series_a, series_b):
        assert isinstance(series_a, (pd.Series, pl.Series))
        assert isinstance(series_b, (pd.Series, pl.Series))
        assert len(series_a) == len(series_b)
        base_type_a = _series_element_classname(series_a)
        base_type_b = _series_element_classname(series_b)
        if type(series_a) == type(series_b):
            assert base_type_a == base_type_b
        assert (len(series_a)-len(_series_drop_nans(series_a)) > 0) == (len(series_b) - len(_series_drop_nans(series_b)) > 0)

    assert isinstance(series, (pd.Series, pl.Series))

    var = MetaVar.fit(series)
    new_series = var.draw_series(len(series))
    check_similar(series, new_series)
    assert var.var_type == var_type
    assert var_type in var.distribution.var_type
    assert var.creation_method["created_by"] == "metasyn"


    new_var = MetaVar.from_dict(var.to_dict())
    with raises(ValueError):
        var_dict = var.to_dict()
        var_dict.update({"type": "unknown"})
        MetaVar.from_dict(var_dict)

    with raises(ValueError):
        var_dict = var.to_dict()
        var_dict["distribution"].update({"implements": "unknown"})
        MetaVar.from_dict(var_dict)

    newer_series = new_var.draw_series(len(series))
    check_similar(newer_series, series)

    assert type(new_var) == type(var)
    assert new_var.dtype == var.dtype
    assert var_type == new_var.var_type

    # Write to JSON file and read it back
    tmp_fp = temp_path / "tmp.json"
    with open(tmp_fp, "w") as f:
        json.dump(_jsonify(var.to_dict()), f)

    with open(tmp_fp, "r") as f:
        new_var = MetaVar.from_dict(json.load(f))
    check_similar(series, new_var.draw_series(len(series)))

    assert type(new_var) == type(var)
    assert new_var.dtype == var.dtype
    assert new_var.var_type == var_type
    assert new_var.creation_method["created_by"] == "metasyn"

    return new_series


@mark.parametrize(
    "series",
    [
        pd.Series(np.random.choice(["a", "b", "c", None], size=100), dtype="category"),
        pl.Series(np.random.choice(["a", "b", "c", None], size=100).tolist(), dtype=pl.Categorical)
    ]
)
def test_categorical(tmp_path, series):
    new_series = check_var(series, "categorical", tmp_path)
    assert set(_series_drop_nans(series)) == set(np.unique(_series_drop_nans(new_series)))

@mark.parametrize("dtype", ["int8", "int16", "int32", "int64", "int"])
def test_integer(dtype, tmp_path):
    series = pd.Series([np.random.randint(0, 10) for _ in range(300)], dtype=dtype)
    new_series = check_var(series, "discrete", tmp_path)
    assert new_series.min() >= 0
    assert new_series.max() < 10


@mark.parametrize("dtype", ["Int8", "Int16", "Int32", "Int64"])
def test_nullable_integer(dtype, tmp_path):
    series = pd.Series([np.random.randint(0, 10) if np.random.rand() > 0.5 else None
                        for _ in range(500)], dtype=dtype)
    new_series = check_var(series, "discrete", tmp_path)
    assert new_series.min() >= 0
    assert new_series.max() < 10


@mark.parametrize(
    "series_type",
    [pl.Series, pd.Series],
)
def test_float(tmp_path, series_type):
    np.random.seed(3727442)
    series = series_type([np.random.rand() for _ in range(10000)])
    new_series = check_var(series, "continuous", tmp_path)
    assert new_series.min() > 0
    assert new_series.max() < 1

    series = pd.Series(np.random.randn(1000))
    check_var(series, "continuous", tmp_path)


@mark.parametrize(
    "series_type",
    [pl.Series, pd.Series],
)
def test_string(tmp_path, series_type):
    series = series_type(np.random.choice(["a", "b", "c", None], size=100).tolist())
    new_series = check_var(series, "string", tmp_path)
    assert set(np.unique(_series_drop_nans(series))) == set(np.unique(_series_drop_nans(new_series)))


@mark.parametrize(
    "series_type",
    [pl.Series, pd.Series]
)
def test_bool(tmp_path, series_type):
    series = series_type(np.random.choice([True, False], size=100))
    check_var(series, "categorical", tmp_path)
    var = MetaVar.fit(series)
    new_series = var.draw_series(10)
    assert new_series.dtype == pl.Boolean


@mark.parametrize(
    "prop_missing",
    [-1, -0.1, 1.2],
)
def test_invalid_prop(prop_missing):
    # with raises(ValueError):
    MetaVar("test", "discrete", DiscreteUniformDistribution.default_distribution())
    with raises(ValueError):
        MetaVar("test", "discrete", DiscreteUniformDistribution.default_distribution(),
                prop_missing=prop_missing)


@mark.parametrize(
    "series,var_type",
    [
        (pl.Series([1, 2, 3]), "discrete"),
        (pl.Series([1.0, 2.0, 3.0]), "continuous"),
        (pl.Series(["1", "2", "3"]), "string"),
        (pl.Series(["1", "2", "3"], dtype=pl.Categorical), "categorical"),
        (pl.Series([dt.time.fromisoformat("10:38:12"), dt.time.fromisoformat("12:52:11")]),
            "time"),
        (pl.Series([dt.datetime.fromisoformat("2022-07-15T10:39:36"),
                    dt.datetime.fromisoformat("2022-08-15T10:39:36")]),
            "datetime"),
        (pl.Series([dt.date.fromisoformat("1903-07-15"), dt.date.fromisoformat("1940-07-16")]),
            "date"),
    ]
)
def test_get_var_type(series, var_type):
    assert MetaVar.get_var_type(series) == var_type


@mark.parametrize(
    "series",
    [pd.Series([np.random.rand() for _ in range(5000)]),
     pl.Series([np.random.rand() for _ in range(5000)])]
)
def test_manual_fit(series):
    var = MetaVar.fit(series)
    assert isinstance(var.distribution, (UniformDistribution, TruncatedNormalDistribution))
    var = MetaVar.fit(series, dist_spec={"implements": "normal"})
    assert isinstance(var.distribution, NormalDistribution)
    var = MetaVar.fit(series, dist_spec=UniformDistribution)
    assert isinstance(var.distribution, UniformDistribution)
    var = MetaVar.fit(series, dist_spec=NormalDistribution(0, 1))
    assert isinstance(var.distribution, NormalDistribution)
    # with raises(TypeError):
        # var.fit(10)


@mark.parametrize(
    "series",
    [pd.Series([pd.NA for _ in range(10)]),
     pl.Series([None for _ in range(10)])]
)
def test_na_zero(series):
    var = MetaVar.fit(series)
    assert var.var_type == "continuous"
    assert var.prop_missing == 1.0


@mark.parametrize(
    "series",
    [pd.Series(np.array([pd.NA if i != 0 else 1.0 for i in range(10)])),
     pl.Series([None if i != 0 else 1.0 for i in range(10)])]
)
def test_na_one(series):
    var = MetaVar.fit(series)
    assert var.var_type == "continuous"
    assert abs(var.prop_missing-0.9) < 1e7


@mark.parametrize(
    "series",
    [pd.Series(np.array([np.nan if i < 2 else 0.123*i for i in range(10)])),
     pl.Series([None if i < 2 else 0.123*i for i in range(10)])]
)
def test_na_two(series):
    var = MetaVar.fit(series)
    assert var.var_type == "continuous"
    assert abs(var.prop_missing-0.8) < 1e7


@mark.parametrize(
    "series",
    [pd.Series(np.random.randint(0, 100000, size=1000)),
     pl.Series(np.random.randint(0, 100000, size=1000))]
)
def test_manual_unique_integer(series):
    var = MetaVar.fit(series)
    assert isinstance(var.distribution, DiscreteUniformDistribution)
    var = MetaVar.fit(series, dist_spec = {"unique": True})
    assert isinstance(var.distribution, UniqueKeyDistribution)


@mark.parametrize(
    "series",
    [
        pd.Series(["x213", "2dh2", "4k2kk"]),
        pl.Series(["x213", "2dh2", "4k2kk"]),
    ]
)
def test_manual_unique_string(series):
    # series = pd.Series(["x213", "2dh2", "4k2kk"])
    var = MetaVar.fit(series)
    assert isinstance(var.distribution, RegexDistribution)
    var = MetaVar.fit(series, dist_spec={"unique": True})
    assert isinstance(var.distribution, UniqueRegexDistribution)


@mark.parametrize(
    "series",
    [
        pl.Series([2]*100 + [4]*100 + [12394]*100),
        pl.Series([-123]*12 + [123]*12),
        # pl.Series(["2"]*100 + ["5"]*100 + ["123"]*100),
        pl.Series(["2"]*100 + ["5"]*100 + ["123"]*100, dtype=pl.Categorical)
    ]
)
def test_int_multinoulli(series):
    var = MetaVar.fit(series)
    assert isinstance(var.distribution, MultinoulliDistribution)


def test_error_multinoulli():
    with raises(ValueError):
        MultinoulliDistribution(["1", "2"], [-0.1, 1.1])
    with pytest.warns():
        MultinoulliDistribution(["1", "2"], [0.1, 0.2])



import json
import subprocess
import sys
from pathlib import Path

import jsonschema
import polars as pl
from pytest import fixture, mark

from metasyn import MetaFrame
from metasyn.validation import validate_gmf_dict

TMP_DIR_PATH = None


@fixture(scope="module")
def tmp_dir(tmp_path_factory) -> Path:
    global TMP_DIR_PATH
    if TMP_DIR_PATH is None:
        # Create a temporary input file
        TMP_DIR_PATH = tmp_path_factory.mktemp("data")
        json_path = TMP_DIR_PATH / "titanic.json"
        csv_fp = Path("tests", "data", "titanic.csv")
        csv_dt = {
            "PassengerId": pl.Int64,
            "Survived": pl.Categorical,
            "Pclass": pl.Categorical,
            "Name": str,
            "Sex": pl.Categorical,
            "SibSp": pl.Categorical,
            "Parch": pl.Categorical,
            "Ticket": str,
            "Cabin": str,
            "Embarked": pl.Categorical,
            "Age": float,
            "Fare": float
        }
        data_frame = pl.read_csv(csv_fp, dtypes=csv_dt)[:100]
        meta_frame = MetaFrame.fit_dataframe(data_frame, var_specs=[{"name": "PassengerId", "distribution": {"unique": True}}])
        meta_frame.to_json(json_path)
        config_fp = TMP_DIR_PATH / "config.ini"
        with open(config_fp, "w") as handle:
            handle.write("""
[[var]]
name = "PassengerId"
distribution = {unique = true}

[[var]]
name = "Fare"
prop_missing = 0.2
distribution = {implements = "lognormal"}
""")
    return TMP_DIR_PATH


@mark.parametrize("ext", [".csv", ".feather", ".parquet", ".pkl", ".xlsx"])
def test_cli(tmp_dir, ext):
    """A simple integration test for reading and writing using the CLI"""

    # create out file path with correct extension
    out_file = tmp_dir / f"titanic{ext}"

    # create command to run in subprocess with arguments
    cmd = [
        Path(sys.executable).resolve(),   # the python executable
        Path("metasyn", "__main__.py"),   # the cli script
        "synthesize",                     # the subcommand
        "-n 25",                          # only generate 25 samples
        tmp_dir / "titanic.json",         # the input file
        "-o",
        out_file                          # the output file
    ]

    # Run the cli with different extensions
    result = subprocess.run(cmd, check=False)
    assert result.returncode == 0, (result.stdout, result.stderr)
    assert out_file.is_file()
    if ext == ".csv":
        df = pl.read_csv(out_file)
        assert len(df) == 25


@mark.parametrize("config", [True, False])
def test_create_meta(tmp_dir, config):
    out_file = tmp_dir / "test.json"
    cmd = [
        Path(sys.executable).resolve(),     # the python executable
        Path("metasyn", "__main__.py"),     # the cli script
        "create-meta",                      # the subcommand
        Path("tests", "data", "titanic.csv"),  # the input file
        "-o",
        out_file                            # the output file
    ]
    if config:
        cmd.extend(["--config", Path(tmp_dir) / 'config.ini'])
    result = subprocess.run(cmd, check=False, capture_output=True)
    assert result.returncode == 0
    assert out_file.is_file()
    meta_frame = MetaFrame.from_json(out_file)
    assert len(meta_frame.meta_vars) == 12


def test_schema_list():
    cmd = [
        Path(sys.executable).resolve(),     # the python executable
        Path("metasyn", "__main__.py"),     # the cli script
        "schema",
        "--list"
    ]
    result = subprocess.run(cmd, check=False, capture_output=True)
    assert result.returncode == 0
    assert "builtin" in result.stdout.decode()


def test_schema_gen(tmp_dir):
    titanic_json = tmp_dir / "titanic.json"
    cmd = [
        Path(sys.executable).resolve(),     # the python executable
        Path("metasyn", "__main__.py"),     # the cli script
        "schema",
        "builtin"
    ]
    result = subprocess.run(cmd, check=False, capture_output=True)
    assert result.returncode == 0
    json_schema = json.loads(result.stdout.decode())
    with open(titanic_json, "r") as handle:
        gmf_dict = json.load(handle)
    validate_gmf_dict(gmf_dict)
    jsonschema.validate(gmf_dict, json_schema)

    cmd.append("non-existent-plugin")
    result = subprocess.run(cmd, check=False, capture_output=True)
    assert result.returncode != 0


def test_datafree(tmp_dir):
    gmf_fp = tmp_dir / "gmf_out.json"
    syn_fp = tmp_dir / "test_out.csv"
    cmd = [
        Path(sys.executable).resolve(),     # the python executable
        Path("metasyn", "__main__.py"),     # the cli script
        "create-meta",                      # the subcommand
        "--config", Path("tests", "data", "no_data_config.toml"),
        "--output", gmf_fp,              # the output file
    ]
    result = subprocess.run(cmd, check=False, capture_output=True)
    assert result.returncode == 0
    meta_frame = MetaFrame.from_json(gmf_fp)
    assert meta_frame.n_rows == 100
    assert len(meta_frame.meta_vars) == 3
    cmd2 = [
        Path(sys.executable).resolve(),     # the python executable
        Path("metasyn", "__main__.py"),     # the cli script
        "synthesize",
        gmf_fp,
        "-o",
        syn_fp
    ]
    result = subprocess.run(cmd2, check=False, capture_output=True)
    assert result.returncode == 0
    df = pl.read_csv(syn_fp)
    assert list(df.columns) == ["PassengerId", "Name", "Cabin"]
    assert len(df) == 100



from math import fabs

import numpy as np
import pandas as pd
import polars as pl
from pytest import mark
from scipy import stats

from metasyn.distribution.discrete import (
    DiscreteNormalDistribution,
    DiscreteTruncatedNormalDistribution,
    DiscreteUniformDistribution,
    PoissonDistribution,
    UniqueKeyDistribution,
)


@mark.parametrize(
    "data",
    [
        [1, 2, 3, 4, 5],
        [-399, 12, 1, 492],
        [np.random.randint(0, 1000) for _ in range(1000)],
    ]
)
@mark.parametrize(
    "series_type", [pd.Series, pl.Series]
)
def test_uniform(data, series_type):
    series = series_type(data)
    dist = DiscreteUniformDistribution.fit(series)
    assert dist.lower == series.min() and dist.upper == series.max()+1
    drawn_values = set([dist.draw() for _ in range(1000)])
    if dist.upper - dist.lower < 5:
        assert len(drawn_values) == len(series)
    drawn_values = np.array(list(drawn_values))
    assert np.isclose(dist.information_criterion(drawn_values),
                      np.log(len(drawn_values))*2+2*len(drawn_values)*(np.log(dist.upper-dist.lower)))


@mark.parametrize(
    "data,better_than_uniform,consecutive",
    [
        ([1, 2, 3, 4, 5], True, True),
        ([5, 4, 3, 2, 1], True, False),
        ([2, 4, 5, 7, 10, 6], True, False),
        ([-3, 1, -5, 3, -2, 0], True, False),
        ([-129384, 2198384, 293, 1293840], False, False),
        ([1, 1, 2, 2, 3, 3], False, False)
    ]
)
@mark.parametrize("series_type", [pd.Series, pl.Series])
def test_integer_key(data, better_than_uniform, consecutive, series_type):
    series = series_type(data)
    dist = UniqueKeyDistribution.fit(series)
    unif_dist = DiscreteUniformDistribution.fit(series)
    assert dist.lower == series.min()
    assert dist.consecutive == consecutive
    assert better_than_uniform == (dist.information_criterion(series) < unif_dist.information_criterion(series))
    assert isinstance(dist, UniqueKeyDistribution)

    drawn_values = np.array([dist.draw() for _ in range(100)])
    if consecutive:
        assert np.all(drawn_values == np.arange(dist.lower, dist.lower+100))


@mark.parametrize("series_type", [pd.Series, pl.Series])
def test_poisson(series_type):
    series = series_type(stats.poisson(mu=10).rvs(1000))
    dist = PoissonDistribution.fit(series)
    dist_unif = DiscreteUniformDistribution.fit(series)
    assert fabs(dist.rate - 10) < 1
    assert dist.information_criterion(series) < dist_unif.information_criterion(series)

@mark.parametrize(
    "lower,upper,mean,sd",
    [
        (-1, 1, 0, 0.5),
        (-10, -8, 0, 5),
        (-5, 3, 2, 2),
    ]
)
def test_trunc_normal(lower, upper, mean, sd):
    a, b = (lower-mean)/sd, (upper-mean)/sd
    values = pl.Series(stats.truncnorm(a=a, b=b, loc=mean, scale=sd).rvs(5000)).cast(pl.Int64)
    dist = DiscreteTruncatedNormalDistribution.fit(values)
    dist_uniform = DiscreteUniformDistribution.fit(values)
    assert dist.information_criterion(values) < dist_uniform.information_criterion(values)
    assert isinstance(dist.draw(), int)


@mark.parametrize(
    "mean,sd",
    [
        (0, 1),
        (-10, 2),
        (100, 200),
    ]
)
def test_normal(mean, sd):
    values = pl.Series(stats.norm(loc=mean, scale=sd).rvs(1000)).cast(pl.Int64)
    dist = DiscreteNormalDistribution.fit(values)
    dist_uniform = DiscreteUniformDistribution.fit(values)
    assert dist.information_criterion(values) < dist_uniform.information_criterion(values)
    assert (dist.mean - mean)/sd < 0.5
    assert (dist.sd - sd)/sd < 0.5
    assert isinstance(dist.draw(), int)



from pytest import mark, raises

from metasyn.distribution.categorical import MultinoulliDistribution
from metasyn.distribution.continuous import NormalDistribution, UniformDistribution
from metasyn.distribution.discrete import DiscreteNormalDistribution, DiscreteUniformDistribution
from metasyn.distribution.string import (
    FakerDistribution,
    RegexDistribution,
    UniqueRegexDistribution,
)
from metasyn.provider import DistributionProviderList


@mark.parametrize(
    "dist_str,var_type,is_unique,dist",
    [
        ("multinoulli", "string", False, MultinoulliDistribution),
        ("multinoulli", "discrete", False, MultinoulliDistribution),
        ("uniform", "continuous", False, UniformDistribution),
        ("uniform", "discrete", False, DiscreteUniformDistribution),
        ("normal", "continuous", False, NormalDistribution),
        ("normal", "discrete", False, DiscreteNormalDistribution),
        ("faker", "string", False, FakerDistribution),
        ("regex", "string", False, RegexDistribution),
        ("regex", "string", True, UniqueRegexDistribution),
    ]
)
def test_util(dist_str, var_type, is_unique, dist):
    provider_list = DistributionProviderList("builtin")
    dist_class = provider_list.find_distribution(dist_str, var_type=var_type, unique=is_unique)
    assert dist == dist_class
    if "faker" in dist_str:
        with raises(ValueError):
            provider_list.find_distribution("this is not a distribution", "string")
    new_class = provider_list.find_distribution(dist_class.__name__, var_type=var_type, unique=is_unique)
    assert new_class == dist_class



"""Package providing different distributions used in metasyn.

The package consists of several distribution modules, it also includes
``base`` module which forms the basis of all distributions.

Each distribution class provides methods for fitting the distribution to a
a series of values, and for generating synthetic data based on the fitted
distribution. Each distribution class also provides a way to calculate the
information criterion, used for selecting the best distribution for
a given set of values.
"""  # pylint: disable=invalid-name

from metasyn.distribution.categorical import MultinoulliDistribution
from metasyn.distribution.continuous import (
    ConstantDistribution,
    ExponentialDistribution,
    LogNormalDistribution,
    NormalDistribution,
    TruncatedNormalDistribution,
    UniformDistribution,
)
from metasyn.distribution.datetime import (
    DateConstantDistribution,
    DateTimeConstantDistribution,
    DateTimeUniformDistribution,
    DateUniformDistribution,
    TimeConstantDistribution,
    TimeUniformDistribution,
)
from metasyn.distribution.discrete import (
    DiscreteConstantDistribution,
    DiscreteNormalDistribution,
    DiscreteTruncatedNormalDistribution,
    DiscreteUniformDistribution,
    PoissonDistribution,
    UniqueKeyDistribution,
)
from metasyn.distribution.na import NADistribution
from metasyn.distribution.string import (
    FakerDistribution,
    FreeTextDistribution,
    RegexDistribution,
    StringConstantDistribution,
    UniqueFakerDistribution,
    UniqueRegexDistribution,
)

__all__ = [
    "MultinoulliDistribution",
    "ExponentialDistribution",
    "LogNormalDistribution",
    "NormalDistribution",
    "TruncatedNormalDistribution",
    "UniformDistribution",
    "DateUniformDistribution",
    "DateTimeUniformDistribution",
    "TimeUniformDistribution",
    "DiscreteNormalDistribution",
    "DiscreteTruncatedNormalDistribution",
    "DiscreteUniformDistribution",
    "PoissonDistribution",
    "UniqueKeyDistribution",
    "FakerDistribution",
    "FreeTextDistribution",
    "UniqueFakerDistribution",
    "NADistribution",
    "RegexDistribution",
    "UniqueRegexDistribution",
    "ConstantDistribution",
    "DiscreteConstantDistribution",
    "StringConstantDistribution",
    "DateConstantDistribution",
    "TimeConstantDistribution",
    "DateTimeConstantDistribution",
]



"""Module implementing distribution providers.

Distribution providers are used to find/fit distributions that are available.
See pyproject.toml on how the builtin distribution provider is registered.
"""

from __future__ import annotations

import warnings
from abc import ABC
from typing import TYPE_CHECKING, Any, List, Optional, Type, Union

try:
    from importlib_metadata import EntryPoint, entry_points
except ImportError:
    from importlib.metadata import EntryPoint, entry_points  # type: ignore

import numpy as np
import polars as pl

from metasyn.distribution.base import BaseDistribution
from metasyn.distribution.categorical import MultinoulliDistribution
from metasyn.distribution.continuous import (
    ConstantDistribution,
    ExponentialDistribution,
    LogNormalDistribution,
    NormalDistribution,
    TruncatedNormalDistribution,
    UniformDistribution,
)
from metasyn.distribution.datetime import (
    DateConstantDistribution,
    DateTimeConstantDistribution,
    DateTimeUniformDistribution,
    DateUniformDistribution,
    TimeConstantDistribution,
    TimeUniformDistribution,
)
from metasyn.distribution.discrete import (
    DiscreteConstantDistribution,
    DiscreteNormalDistribution,
    DiscreteTruncatedNormalDistribution,
    DiscreteUniformDistribution,
    PoissonDistribution,
    UniqueKeyDistribution,
)
from metasyn.distribution.na import NADistribution
from metasyn.distribution.string import (
    FakerDistribution,
    FreeTextDistribution,
    RegexDistribution,
    StringConstantDistribution,
    UniqueFakerDistribution,
    UniqueRegexDistribution,
)
from metasyn.privacy import BasePrivacy, BasicPrivacy
from metasyn.util import get_registry
from metasyn.varspec import DistributionSpec

if TYPE_CHECKING:
    from metasyn.config import VarSpec, VarSpecAccess


class BaseDistributionProvider(ABC):
    """Base class for all distribution providers.

    A distribution provider is a class that provides a set of distributions
    that can be used by metasyn to generate synthetic data.
    This class acts as a base class for creating specific distribution
    providers. It also contains a list of the available distributions and
    legacy distributions. A list of distributions for a specific type can be
    accessed with ``get_dist_list``.
    """

    name = ""
    version = ""
    distributions: list[type[BaseDistribution]] = []
    legacy_distributions: list[type[BaseDistribution]] = []

    def __init__(self):
        # Perform internal consistency check.
        assert len(self.name) > 0
        assert len(self.version) > 0
        assert len(self.distributions) > 0

    def get_dist_list(self, var_type: Optional[str],
                      unique: bool = False,
                      use_legacy: bool = False) -> List[Type[BaseDistribution]]:
        """Get all distributions for a certain variable type.

        Parameters
        ----------
        var_type:
            Variable type to get the distributions for.
        unique:
            Whether the distirbutions should be unique.
        use_legacy:
            Whether to find the distributions in the legacy distribution list.

        Returns
        -------
        list[Type[BaseDistribution]]:
            List of distributions with that variable type.
        """
        dist_list = []
        if use_legacy:
            distributions = self.legacy_distributions
        else:
            distributions = self.distributions
        distributions = [d for d in distributions if d.unique == unique]
        if var_type is None:
            return distributions
        for dist_class in distributions:
            str_chk = (isinstance(dist_class.var_type, str) and var_type == dist_class.var_type)
            lst_chk = (not isinstance(dist_class.var_type, str) and var_type in dist_class.var_type)
            if str_chk or lst_chk:
                dist_list.append(dist_class)
        return dist_list

    @property
    def all_var_types(self) -> List[str]:
        """Return list of available variable types."""
        var_type_set = set()
        for dist in self.distributions:
            if isinstance(dist.var_type, str):
                var_type_set.add(dist.var_type)
            else:
                var_type_set.update(dist.var_type)
        return list(var_type_set)


class BuiltinDistributionProvider(BaseDistributionProvider):
    """Distribution tree that includes the builtin distributions.

    This class inherits from BaseDistributionProvider and provides
    the built-in metasyn distributions.
    """

    name = "builtin"
    version = "1.2"
    distributions = [
        DiscreteNormalDistribution, DiscreteTruncatedNormalDistribution,
        DiscreteUniformDistribution, PoissonDistribution, UniqueKeyDistribution,
        UniformDistribution, NormalDistribution, LogNormalDistribution,
        TruncatedNormalDistribution, ExponentialDistribution,
        MultinoulliDistribution,
        RegexDistribution, UniqueRegexDistribution, FakerDistribution, UniqueFakerDistribution,
        FreeTextDistribution,
        DateUniformDistribution,
        TimeUniformDistribution,
        DateTimeUniformDistribution,
        ConstantDistribution,
        DiscreteConstantDistribution,
        StringConstantDistribution,
        DateTimeConstantDistribution,
        DateConstantDistribution,
        TimeConstantDistribution,
    ]
    legacy_distributions = []


class DistributionProviderList():
    """List of DistributionProviders with functionality to fit distributions.

    This class is responsible for managing and providing access to
    different distribution providers. It allows for fitting distributions,
    as well as retrieving distributions based on certain constraints
    such as privacy level, variable type, and uniqueness.

    Parameters
    ----------
    dist_providers:
        One or more distribution providers, that are denoted either with a string ("builtin"),
        DistributionProvider (BuiltinDistributionProvider())
        or DistributionProvider type (BuiltinDistributionProvider).
        The order in which distribution providers are included matters.
        If a provider implements the same distribution at the same privacy level,
        then only the first will be taken into account.
    """

    def __init__(
            self,
            dist_providers: Union[
                list[str],
                None, str, type[BaseDistributionProvider], BaseDistributionProvider,
                list[Union[str, type[BaseDistributionProvider], BaseDistributionProvider]]]):
        if dist_providers is None:
            self.dist_packages = _get_all_provider_list()
            return

        if isinstance(dist_providers, (str, type, BaseDistributionProvider)):
            dist_providers = [dist_providers]
        self.dist_packages = []
        for provider in dist_providers:
            if isinstance(provider, str):
                self.dist_packages.append(get_distribution_provider(provider))
            elif isinstance(provider, type):
                self.dist_packages.append(provider())
            elif isinstance(provider, BaseDistributionProvider):
                self.dist_packages.append(provider)
            else:
                raise ValueError(f"Unknown distribution package type '{type(provider)}'")

    def fit(self, series: pl.Series,
            var_type: str,
            dist_spec: DistributionSpec,
            privacy: BasePrivacy = BasicPrivacy()) -> BaseDistribution:
        """Fit a distribution to a column/series.

        Parameters
        ----------
        series:
            The data to fit the distributions to.
        var_type:
            The variable type of the data.
        dist_spec:
            Distribution to fit. If not supplied or None, the information
            criterion will be used to determine which distribution is the most
            suitable. For most variable types, the information criterion is based on
            the BIC (Bayesian Information Criterion).
        privacy:
            Level of privacy that will be used in the fit.
        """
        if dist_spec.distribution is not None:
            return dist_spec.distribution
        if dist_spec.implements is not None:
            return self._fit_distribution(series, dist_spec, var_type, privacy)
        return self._find_best_fit(series, var_type, dist_spec.unique, privacy)

    def create(self, var_spec: Union[VarSpec, VarSpecAccess]) -> BaseDistribution:
        """Create a distribution without any data.

        Parameters
        ----------
        var_spec
            A variable configuration that provides all the qinformation to create the distribution.

        Returns
        -------
            A distribution according to the variable specifications.
        """
        dist_spec = var_spec.dist_spec
        unique = dist_spec.unique if dist_spec.unique else False
        assert dist_spec.implements is not None and var_spec.var_type is not None
        dist_class = self.find_distribution(
            dist_spec.implements, var_spec.var_type,
            privacy=BasicPrivacy(), unique=unique)
        return dist_class(**dist_spec.parameters)

    def _find_best_fit(self, series: pl.Series, var_type: str,
                       unique: Optional[bool],
                       privacy: BasePrivacy) -> BaseDistribution:
        """Fit a distribution to a series.

        Search for the distribution within all available distributions in the tree.

        Parameters
        ----------
        series:
            Series to fit a distribution to.
        var_type:
            Variable type of the series.
        unique:
            Whether the variable should be unique or not.
        privacy:
            Privacy level to find the best fit with.

        Returns
        -------
        BaseDistribution:
            Distribution fitted to the series.
        """
        if len(series.drop_nulls()) == 0:
            return NADistribution()
        try_unique = unique if unique is True else False
        dist_list = self.get_distributions(privacy, var_type, unique=try_unique)
        if len(dist_list) == 0:
            raise ValueError(f"No available distributions with variable type: '{var_type}'"
                             f" and unique={try_unique}")
        dist_instances = [d.fit(series, **privacy.fit_kwargs) for d in dist_list]
        dist_bic = [d.information_criterion(series) for d in dist_instances]
        if unique is None:
            dist_list_unq = self.get_distributions(privacy, var_type, unique=True)
            if len(dist_list_unq) > 0:
                dist_inst_unq = [d.fit(series, **privacy.fit_kwargs) for d in dist_list_unq]
                dist_bic_unq = [d.information_criterion(series) for d in dist_inst_unq]
                if np.min(dist_bic_unq) < np.min(dist_bic):
                    warnings.warn(
                        f"\nVariable '{series.name}' was detected to be unique, but has not"
                        f" explicitly been set to unique.\n"
                        f"To generate only unique values for column '{series.name}', "
                        f"set unique to True.\n"
                        f"To dismiss this warning, set unique to False.",
                        UserWarning
                    )

        return dist_instances[np.argmin(dist_bic)]

    def find_distribution(self,  # pylint: disable=too-many-branches
                          dist_name: str,
                          var_type: str,
                          privacy: Optional[BasePrivacy] = BasicPrivacy(),
                          unique: bool = False,
                          version: Optional[str] = None) -> type[BaseDistribution]:
        """Find a distribution and fit keyword arguments from a name.

        Parameters
        ----------
        dist_name:
            Name of the distribution, e.g., for the built-in
            uniform distribution: "uniform", "core.uniform", "UniformDistribution".
        privacy:
            Type of privacy to be applied.
        var_type:
            Type of the variable to find. If var_type is None, then do not check the
            variable type.
        unique:
            Whether the distribution to be found is unique.
        version:
            Version of the distribution to get. If necessary get them from legacy.

        Returns
        -------
        tuple[Type[BaseDistribution], dict[str, Any]]:
            A distribution and the arguments to create an instance.
        """
        if NADistribution.matches_name(dist_name):
            return NADistribution

        versions_found = []
        for dist_class in self.get_distributions(
                privacy, var_type=var_type, unique=unique) + [NADistribution]:
            if dist_class.matches_name(dist_name):
                if version is None or version == dist_class.version:
                    return dist_class
                versions_found.append(dist_class)

        # Look for distribution in legacy
        warnings.simplefilter("always")
        legacy_distribs = [
            dist_class
            for dist_class in self.get_distributions(privacy, use_legacy=True,
                                                     var_type=var_type, unique=unique)
            if dist_class.matches_name(dist_name)]

        if len(legacy_distribs + versions_found) == 0:
            registry = get_registry()
            available = {}
            for plugin, meta in registry.items():
                if dist_name in meta["distributions"]:
                    available[plugin] = meta["url"]
            if len(available) > 0:
                avail_str = "\n".join("{plugin}: {url}" for plugin, url in available.items())
                raise ValueError(f"You are trying to use a distribution named '{dist_name}', \n"
                                 f"but it is not installed.\n"
                                 f"\n"
                                 f"{dist_name} is available from:\n\n{avail_str}\n")
            raise ValueError(f"Cannot find distribution with name '{dist_name}'.")

        if len(versions_found) == 0:
            warnings.warn("Distribution with name '{dist_name}' is deprecated and "
                          "will be removed in the future.")

        # Find exact matches in legacy distributions
        legacy_versions = [dist.version for dist in legacy_distribs]
        if version is not None and version in legacy_versions:
            warnings.warn("Version ({version}) of distribution with name '{dist_name}'"
                          " is deprecated and will be removed in the future.")
            return legacy_distribs[legacy_versions.index(version)]

        # If version is None, take the latest version.
        if version is None:
            scores = [int(dist.version.split(".")[0]) * 100 + int(dist.version.split(".")[1])
                      for dist in legacy_distribs]
            i_min = np.argmax(scores)
            return legacy_distribs[i_min]

        # Find the distribution with the same major revision, and closest minor revision.
        major_version = int(version.split(".")[0])
        minor_version = int(version.split(".")[1])
        all_dist = versions_found + legacy_distribs
        all_versions = [[int(x) for x in dist.version.split(".")] for dist in all_dist]
        score = [int(ver[0] == major_version) * 1000000 - (ver[1] - minor_version) ** 2
                 for ver in all_versions]
        i_max = np.argmax(score)

        # Wrong major revision.
        if score[i_max] < 500000:
            raise ValueError(
                f"Cannot find compatible version for distribution '{dist_name}', available: "
                f"{legacy_versions + versions_found}")

        warnings.warn("Version mismatch ({version}) versus ({all_dist[i_max].version}))")
        return all_dist[i_max]

    def _fit_distribution(self, series: pl.Series,
                          dist_spec: DistributionSpec,
                          var_type: str,
                          privacy: BasePrivacy) -> BaseDistribution:
        """Fit a specific distribution to a series.

        In contrast the fit method, this needs a supplied distribution(type).

        Parameters
        ----------
        series:
            Series to fit the distribution to.
        dist_spec:
            Distribution to fit (if it is not already fitted).
        var_type:
            Type of variable to fit the distribution for.
        privacy:
            Privacy level to fit the distribution with.

        Returns
        -------
        BaseDistribution:
            Fitted distribution.
        """
        unique = dist_spec.unique
        unique = unique if unique else False
        assert dist_spec.implements is not None

        # If the parameters are already specified, the privacy level doesn't matter anymore.
        if dist_spec.parameters is not None:
            dist_class = self.find_distribution(dist_spec.implements, var_type, unique=unique)
            return dist_class(**dist_spec.parameters)

        dist_class = self.find_distribution(dist_spec.implements, var_type, privacy=privacy,
                                            unique=unique)

        if issubclass(dist_class, NADistribution):
            dist_instance = dist_class.default_distribution()
        else:
            fit_kwargs = dist_spec.fit_kwargs
            dist_instance = dist_class.fit(series, **privacy.fit_kwargs, **fit_kwargs)
        return dist_instance

    def get_distributions(self, privacy: Optional[BasePrivacy] = None,
                          var_type: Optional[str] = None,
                          unique: bool = False,
                          use_legacy: bool = False) -> list[type[BaseDistribution]]:
        """Get the available distributions with constraints.

        Parameters
        ----------
        privacy:
            Privacy level/type to filter the distributions.
        var_type:
            Variable type to filter for, e.g. 'string'.
        unique:
            Whether the distributions to be gotten are unique.
        use_legacy:
            Whether to use legacy distributions or not.

        Returns
        -------
        dist_list:
            List of distributions that fit the given constraints.
        """
        dist_list = []
        for dist_provider in self.dist_packages:
            dist_list.extend(dist_provider.get_dist_list(
                var_type, use_legacy=use_legacy, unique=unique))

        if privacy is None:
            return dist_list
        dist_list = [dist for dist in dist_list if dist.privacy == privacy.name]
        return dist_list

    def from_dict(self, var_dict: dict[str, Any]) -> BaseDistribution:
        """Create a distribution from a dictionary.

        Parameters
        ----------
        var_dict:
            Variable dictionary that includes the distribution properties.

        Returns
        -------
        BaseDistribution:
            Distribution representing the dictionary.
        """
        dist_name = var_dict["distribution"]["implements"]
        version = var_dict["distribution"].get("version", "1.0")
        var_type = var_dict["type"]
        unique = var_dict["distribution"]["unique"]
        dist_class = self.find_distribution(dist_name, version=version,
                                            var_type=var_type, unique=unique)
        return dist_class.from_dict(var_dict["distribution"])


def _get_all_providers() -> dict[str, EntryPoint]:
    """Get all available providers."""
    return {
        entry.name: entry
        for entry in entry_points(group="metasyn.distribution_provider")
    }


def _get_all_provider_list() -> list[BaseDistributionProvider]:
    return [p.load()() for p in _get_all_providers().values()]


def get_distribution_provider(provider: Union[str, type[
                                        BaseDistributionProvider],
                                        BaseDistributionProvider] = "builtin"
                              ) -> BaseDistributionProvider:
    """Get a distribution tree.

    Parameters
    ----------
    provider:
        Name, class or class type of the provider to be used.
    kwargs:
        Extra keyword arguments for initialization of the distribution provider.

    Returns
    -------
    BaseDistributionProvider:
        The distribution provider that was found.
    """
    if isinstance(provider, BaseDistributionProvider):
        return provider
    if isinstance(provider, type):
        return provider()

    all_providers = _get_all_providers()
    try:
        return all_providers[provider].load()()
    except KeyError as exc:
        registry = get_registry()
        if provider not in registry:
            raise ValueError(f"Cannot find distribution provider with name '{provider}'.") from exc
        raise ValueError(f"Distribution provider '{provider}' is not installed.\n"
                         f"See {registry['provider']['url']} for installation instructions."
                         ) from exc



from pathlib import Path
from random import random

import pandas as pd
import polars as pl
import pytest
from pytest import mark

from metasyn.metaframe import MetaFrame
from metasyn.provider import get_distribution_provider
from metasyn.var import MetaVar

dtypes = {
    "PassengerId": "int",
    "Survived": "category",
    "Pclass": "category",
    "Name": "string",
    "Sex": "category",
    "SibSp": "category",
    "Parch": "category",
    "Ticket": "string",
    "Cabin": "string",
    "Embarked": "category",
    "Age": "float",
    "Fare": "float",
}


def _read_csv(fp, dataframe_lib):
    if dataframe_lib == "pandas":
        df = pd.read_csv(fp, dtype=dtypes)
        return df.iloc[:100]
    else:
        df = pl.read_csv(fp, dtypes={x: pl.Categorical for x, x_type in dtypes.items() if x_type == "category"})
        return df[:100]


@mark.parametrize("dataframe_lib", ["polars", "pandas"])
def test_dataset(tmp_path, dataframe_lib):
    titanic_fp = Path("tests", "data", "titanic.csv")
    tmp_fp = tmp_path / "tmp.json"
    df = _read_csv(titanic_fp, dataframe_lib)
    dataset = MetaFrame.fit_dataframe(
        df,
        var_specs=[
                {"name": "Name", "prop_missing": 0.5},
                {"name": "Ticket", "description": "test_description"},
                {"name": "Fare", "distribution": {"implements": "normal"}},
                {"name": "PassengerId", "distribution": {"unique": True}},
             ])

    def check_dataset(dataset):
        assert dataset.n_columns == 12
        assert dataset.n_rows == 100
        sub_types = {
            "int": "discrete",
            "string": "string",
            "category": "categorical",
            "float": "continuous",
        }
        for col, type_name in dtypes.items():
            meta_var_type = sub_types[type_name]
            assert dataset[col].var_type == meta_var_type
        with pytest.raises(KeyError):
            dataset["unknown"]
        with pytest.raises(IndexError):
            dataset[20]
        with pytest.raises(TypeError):
            dataset[MetaVar]

        assert "Rows" in str(dataset)
        assert len(dataset.synthesize()) == dataset.n_rows
        assert len(dataset.synthesize(5)) == 5
        with pytest.raises(ValueError):
            dataset.n_rows = None
            dataset.synthesize()
        dataset.n_rows = 100

    check_dataset(dataset)
    dataset.to_json(tmp_fp)
    dataset = MetaFrame.from_json(tmp_fp)
    check_dataset(dataset)

    dataset.descriptions = {"Embarked": "Some description", "Sex": "Other description"}
    assert dataset.descriptions["Embarked"] == "Some description"
    with pytest.raises(AssertionError):
        dataset.descriptions = "12345"
    dataset.descriptions = {"Embarked": "New description"}
    assert dataset.descriptions["Embarked"] == "New description"
    assert dataset.descriptions["Sex"] == "Other description"
    if isinstance(df, pl.DataFrame):
        col_names = df.columns
    else:
        col_names = list(df)
    dataset.descriptions = col_names
    for name in col_names:
        print(name, dataset.descriptions[name])
        assert dataset.descriptions[name] == name

    assert isinstance(repr(dataset), str)

    # Check whether non-columns raise an error
    with pytest.raises(ValueError):
        dataset = MetaFrame.fit_dataframe(df, var_specs=[{"name": "unicorn", "prop_missing": 0.5}])


def test_distributions(tmp_path):
    tmp_fp = tmp_path / "tmp.json"

    provider = get_distribution_provider()
    for var_type in provider.all_var_types:
        for dist in provider.get_dist_list(var_type):
            var = MetaVar(name="None", var_type=var_type, distribution=dist.default_distribution(),
                          prop_missing=random())
            dataset = MetaFrame([var], n_rows=10)
            dataset.to_json(tmp_fp)



import pandas as pd
import polars as pl
from pytest import mark

from metasyn.distribution.string import FakerDistribution, FreeTextDistribution


@mark.parametrize("series_type", [pd.Series, pl.Series])
def test_faker(series_type):
    var = FakerDistribution.fit(series_type([1, 2, 3]))
    assert isinstance(var.to_dict(), dict)
    assert isinstance(var.draw(), str)
    assert 'faker' in str(var)
    assert var.locale == "en_US"
    assert var.faker_type == "city"


@mark.parametrize(
    "series,lang,avg_sentences,avg_words", [
            (pl.Series(
                ["hotdog", "mother", "yes", "tip", "crate",
                 "sink", "dark", "crossbar", "toilet", "grow", "patient"]), "EN", None, 1),
            (pl.Series(["hotdog mother", "yes tip", "crate sink", "dark crossbar", "toilet grow",
                        "patient is"]), "EN", None, 2),
            (pl.Series(["hotdog mother.", "yes tip.", "crate sink.", "dark crossbar.", "toilet grow.",
                        "patient is."]), "EN", 1, 2),
            (pl.Series(["gaat naar school. Ik ben benieuwd.", "Wat is vraag? Wanneer komt hij."]),
             "NL", 2, 6)
    ])
def test_free_text(series, lang, avg_sentences, avg_words):
    dist = FreeTextDistribution.fit(series)
    assert dist.locale == lang
    assert dist.avg_sentences == avg_sentences
    assert dist.avg_words == avg_words
    dist.draw()



"""Utility module for metasyn."""
from __future__ import annotations

try:
    import tomllib
except ImportError:
    import tomli as tomllib  # type: ignore  # noqa

try:
    from importlib_resources import files
except ImportError:
    from importlib.resources import files  # type: ignore


def get_registry() -> dict:
    """Get the registry dictionary from the package.

    This registry contains information on plugins that are available for metasyn.

    Returns
    -------
        Dictionary containing the registry entries.
    """
    registry_fp = files(__package__) / "schema" / "plugin_registry.toml"
    with open(registry_fp, "rb") as handle:
        registry = tomllib.load(handle)
    return registry



"""Create and retrieve demo datasets."""

# import random
from pathlib import Path

import polars as pl

try:
    from importlib_resources import files
except ImportError:
    from importlib.resources import files  # type: ignore

# import numpy as np
# import wget

# from metasyn.distribution.datetime import (
#     DateTimeUniformDistribution,
#     DateUniformDistribution,
#     TimeUniformDistribution,
# )


# def create_titanic_demo(output_fp: Path) -> Path:
#     """Create demo dataset for the titanic dataset.

#     Arguments
#     ---------
#     output_fp:
#         File to write the demonstration table to.

#     Returns
#     -------
#         Output file location.
#     """
#     titanic_fp = Path("titanic.csv")
#     if Path(output_fp).is_file():
#         return output_fp
#     if not titanic_fp.is_file():
#         wget.download(
#             "https://raw.githubusercontent.com/pandas-dev/pandas/main/doc/data/titanic.csv")
#     dframe = pd.read_csv(titanic_fp)
#     np.random.seed(1283742)
#     random.seed(1928374)

#     # Convert Age to a nullable integer.
#     dframe["Age"] = dframe["Age"].round().astype("Int64")

#     # Add a date column.
#     date_dist = DateUniformDistribution.default_distribution()
#     dframe["Birthday"] = [date_dist.draw() if np.random.rand() < 0.9 else pd.NA
#                           for _ in range(len(dframe))]

#     # Add a time column.

#     time_dist = TimeUniformDistribution.default_distribution()
#     dframe["Board time"] = [time_dist.draw() if np.random.rand() < 0.9 else pd.NA
#                             for _ in range(len(dframe))]

#     # Add a datetime column
#     time_dist = DateTimeUniformDistribution.default_distribution()
#     dframe["Married since"] = [time_dist.draw() if np.random.rand() < 0.9 else pd.NA
#                                for _ in range(len(dframe))]

#     dframe["all_NA"] = [pd.NA for _ in range(len(dframe))]
#     # Remove some columns for brevity and write to a file.
#     dframe = dframe.drop(["SibSp", "Pclass", "Survived"], axis=1)
#     dframe.to_csv(output_fp, index=False)
#     return output_fp


def demo_file(name: str = "titanic") -> Path:
    """Get the path for a demo data file.

    There are three options:
        - titanic (Included in pandas, but post-processed to contain more columns)
        - spaceship (CC-BY from https://www.kaggle.com/competitions/spaceship-titanic)
        - fruit (very basic example data from Polars)

    Arguments
    ---------
    name:
        Name of the demo dataset.

    Returns
    -------
        Path to the dataset.

    """
    if name == "titanic":
        return files(__package__) / "demo_titanic.csv"
    if name == "spaceship":
        return files(__package__) / "demo_spaceship.csv"
    if name == "fruit":
        return files(__package__) / "demo_fruit.csv"

    raise ValueError(
        f"No demonstration dataset with name '{name}'. Options: titanic, spaceship, fruit."
    )


def demo_dataframe(name: str = "titanic") -> pl.DataFrame:
    """Get a demonstration dataset as a prepared polars dataframe.

    There are three options:
        - titanic (Included in pandas, but post-processed to contain more columns)
        - spaceship (CC-BY from https://www.kaggle.com/competitions/spaceship-titanic)
        - fruit (very basic example data from Polars)

    Arguments
    ---------
    name:
        Name of the demo dataset: spaceship, fruit, or titanic.

    Returns
    -------
        Polars dataframe with correct column types

    """
    file_path = demo_file(name=name)
    if name == "spaceship":
        # the Kaggle spaceship data (CC-BY)
        data_types = {
            "HomePlanet": pl.Categorical,
            "CryoSleep": pl.Categorical,
            "VIP": pl.Categorical,
            "Destination": pl.Categorical,
            "Transported": pl.Categorical,
        }
        return pl.read_csv(file_path, dtypes=data_types, try_parse_dates=True)
    if name == "titanic":
        # our edited titanic data
        data_types = {"Sex": pl.Categorical, "Embarked": pl.Categorical}
        return pl.read_csv(file_path, dtypes=data_types, try_parse_dates=True)
    if name == "fruit":
        # basic fruit data from polars example
        data_types = {"fruits": pl.Categorical, "cars": pl.Categorical}
        return pl.read_csv(file_path, dtypes=data_types)

    raise ValueError(
        f"No demonstration dataset with name '{name}'. Options: titanic, spaceship, fruit."
    )



"""Module implementing discrete distributions."""

from typing import Set

import numpy as np
from scipy.stats import poisson, randint

from metasyn.distribution.base import (
    BaseConstantDistribution,
    BaseDistribution,
    ScipyDistribution,
    metadist,
)
from metasyn.distribution.continuous import NormalDistribution, TruncatedNormalDistribution


@metadist(implements="core.uniform", var_type="discrete")
class DiscreteUniformDistribution(ScipyDistribution):
    """Uniform discrete distribution.

    It differs from the floating point uniform distribution by
    being a discrete distribution instead.

    Parameters
    ----------
    lower: int
        Lower bound (inclusive) of the uniform distribution.
    upper: int
        Upper bound (exclusive) of the uniform distribution.

    Examples
    --------
    >>> DiscreteUniformDistribution(lower=3, upper=20)
    """

    dist_class = randint

    def __init__(self, lower: int, upper: int):
        self.par = {"lower": lower, "upper": upper}
        self.dist = self.dist_class(low=lower, high=upper)

    def _information_criterion(self, values):
        return np.log(len(values))*self.n_par - 2*np.sum(self.dist.logpmf(values))

    @classmethod
    def _fit(cls, values):
        param = {"lower": values.min(), "upper": values.max()+1}
        return cls(**param)

    @classmethod
    def default_distribution(cls):
        return cls(0, 10)

    @classmethod
    def _param_schema(cls):
        return {
            "lower": {"type": "integer"},
            "upper": {"type": "integer"},
        }

@metadist(implements="core.normal", var_type="discrete")
class DiscreteNormalDistribution(NormalDistribution):
    """Normal discrete distribution.

    This class implements the normal/gaussian distribution and takes
    the average and standard deviation as initialization input.

    Parameters
    ----------
    mean: float
        Mean of the normal distribution.

    sd: float
        Standard deviation of the normal distribution.

    Examples
    --------
    >>> DiscreteNormalDistribution(mean=2.4, sd=1.2)
    """

    def draw(self):
        return int(super().draw())

@metadist(implements="core.truncated_normal", var_type="discrete")
class DiscreteTruncatedNormalDistribution(TruncatedNormalDistribution):
    """Truncated normal discrete distribution.

    Parameters
    ----------
    lower: float
        Lower bound of the truncated normal distribution.
    upper: float
        Upper bound of the truncated normal distribution.
    mean: float
        Mean of the non-truncated normal distribution.
    sd: float
        Standard deviation of the non-truncated normal distribution.

    Examples
    --------
    >>> DiscreteTruncatedNormalDistribution(lower=1.2, upper=4.5, mean=2.3, sd=4.5)
    """

    def draw(self):
        return int(super().draw())


@metadist(implements="core.poisson", var_type="discrete")
class PoissonDistribution(ScipyDistribution):
    """Poisson distribution.

    Parameters
    ----------
    rate:
        Rate (mean) of the poisson distribution.

    Examples
    --------
    >>> PoissonDistribution(rate=3.5)
    """

    dist_class = poisson

    def __init__(self, rate: float):
        self.par = {"rate": rate}
        self.dist = self.dist_class(mu=rate)

    def _information_criterion(self, values):
        return np.log(len(values))*self.n_par - 2*np.sum(self.dist.logpmf(values))

    @classmethod
    def _fit(cls, values):
        return cls(values.mean())

    @classmethod
    def default_distribution(cls):
        return cls(0.5)

    @classmethod
    def _param_schema(cls):
        return {
            "rate": {"type": "number"},
        }


@metadist(implements="core.unique_key", var_type="discrete", unique=True)
class UniqueKeyDistribution(ScipyDistribution):
    """Unique key distribution for identifiers.

    Discrete distribution that ensures the uniqueness of the drawn values.

    Parameters
    ----------
    lower:
        Minimum value for the keys.
    consecutive:
        True if keys are consecutive and increasing, False otherwise.

    Examples
    --------
    >>> UniqueKeyDistribution(lower=0, consecutive=True)
    """

    def __init__(self, lower: int, consecutive: bool):
        self.par = {"lower": lower, "consecutive": consecutive}
        self.last_key = lower - 1
        self.key_set: Set[int] = set()

    @classmethod
    def _fit(cls, values):
        lower = values.min()
        high = values.max() + 1
        if len(values) == high-lower and np.all(values.to_numpy() == np.arange(lower, high)):
            return cls(lower, True)
        return cls(lower, False)

    def draw_reset(self):
        self.last_key = self.lower - 1
        self.key_set = set()

    def draw(self):
        if self.consecutive == 1:
            self.last_key += 1
            return self.last_key

        while True:
            random_number = np.random.randint(self.lower, self.lower+2*len(self.key_set)+2)
            if random_number not in self.key_set:
                self.key_set.add(random_number)
                return random_number

    def _information_criterion(self, values):
        if values.min() < self.lower:
            return 2*np.log(len(values))+999*len(values)

        # If the values are not unique the fit is extremely bad.
        if len(set(values)) != len(values):
            return 2*np.log(len(values))+999*len(values)

        lower = values.min()
        high = values.max()+1

        if self.consecutive == 1:
            # Check if the values are truly consecutive
            if len(values) == high-lower and np.all(values.to_numpy() == np.arange(lower, high)):
                return 2*np.log(len(values))
            return 2*np.log(len(values))+999*len(values)

        n_choice = high - lower

        # Probabilities go up like 1/n, 1/(n-1), 1/(n-2), ..., 1/2, 1
        return (3*np.log(len(values))
                - 2*np.sum(np.log(1/np.arange(n_choice, n_choice-len(values), -1))))

    @classmethod
    def default_distribution(cls):
        return cls(0, False)

    @classmethod
    def _param_schema(cls):
        return {
            "lower": {"type": "integer"},
            "consecutive": {"type": "boolean"},
        }

@metadist(implements="core.constant", var_type="discrete")
class DiscreteConstantDistribution(BaseConstantDistribution):
    """Constant discrete distribution.

    This class implements the constant distribution, so that it draws always
    the same value.

    Parameters
    ----------
    value: int
        Value that will be returned when drawn.

    Examples
    --------
    >>> DiscreteConstantDistribution(213456)
    """

    @classmethod
    def default_distribution(cls) -> BaseDistribution:
        return cls(0)

    @classmethod
    def _param_schema(cls):
        return {
            "value": {"type": "integer"}
        }



"""Module implementing continuous (floating point) distributions."""

import numpy as np
from scipy.optimize import minimize
from scipy.stats import expon, lognorm, norm, truncnorm, uniform
from scipy.stats._continuous_distns import FitDataError

from metasyn.distribution.base import (
    BaseConstantDistribution,
    BaseDistribution,
    ScipyDistribution,
    metadist,
)


@metadist(implements="core.uniform", var_type="continuous")
class UniformDistribution(ScipyDistribution):
    """Uniform distribution for floating point type.

    This class implements the uniform distribution between a minimum
    and maximum.

    Parameters
    ----------
    lower: float
        Lower bound for uniform distribution.
    upper: float
        Upper bound for uniform distribution.

    Examples
    --------
    >>> UniformDistribution(lower=-3.0, upper=10.0)
    """

    dist_class = uniform

    def __init__(self, lower: float, upper: float):
        self.par = {"lower": lower, "upper": upper}
        self.dist = uniform(loc=self.lower, scale=max(self.upper-self.lower, 1e-8))

    @classmethod
    def _fit(cls, values):
        return cls(values.min(), values.max())

    def _information_criterion(self, values):
        if np.any(np.array(values) < self.lower) or np.any(np.array(values) > self.upper):
            return np.log(len(values))*self.n_par + 100*len(values)
        if np.fabs(self.upper-self.lower) < 1e-8:
            return np.log(len(values))*self.n_par - 100*len(values)
        return (np.log(len(values))*self.n_par
                - 2*len(values)*np.log((self.upper-self.lower)**-1))

    @classmethod
    def default_distribution(cls):
        return cls(0, 1)

    @classmethod
    def _param_schema(cls):
        return {
            "lower": {"type": "number"},
            "upper": {"type": "number"},
        }


@metadist(implements="core.normal", var_type="continuous")
class NormalDistribution(ScipyDistribution):
    """Normal distribution for floating point type.

    This class implements the normal/gaussian distribution and takes
    the average and standard deviation as initialization input.

    Parameters
    ----------
    mean: float
        Mean of the normal distribution.
    sd: float
        Standard deviation of the normal distribution.

    Examples
    --------
    >>> NormalDistribution(mean=1.3, sd=4.5)
    """

    implements = "core.normal"
    dist_class = norm

    def __init__(self, mean: float, sd: float):
        self.par = {"mean": mean, "sd": sd}
        self.dist = norm(loc=mean, scale=max(sd, 1e-8))

    @classmethod
    def default_distribution(cls):
        return cls(0, 1)

    @classmethod
    def _param_schema(cls):
        return {
            "mean": {"type": "number"},
            "sd": {"type": "number"},
        }


@metadist(implements="core.lognormal", var_type="continuous")
class LogNormalDistribution(ScipyDistribution):
    """Log-normal distribution for floating point type.

    This class implements the log-normal mu and sigma as initialization input.

    Parameters
    ----------
    mean: float
        Controls the mean of the distribution.
    sd: float
        Controls the width of the distribution.

    Examples
    --------
    >>> LogNormalDistribution(mean=-2.0, sd=4.5)
    """

    dist_class = lognorm

    def __init__(self, mean: float, sd: float):  # pylint: disable=invalid-name
        self.par = {"mean": mean, "sd": sd}
        self.dist = lognorm(s=max(sd, 1e-8), scale=np.exp(mean))

    @classmethod
    def _fit(cls, values):
        try:
            sd, _, scale = cls.dist_class.fit(values, floc=0)
        except FitDataError:
            return cls(0, 1)
        return cls(np.log(scale), sd)

    @classmethod
    def default_distribution(cls):
        return cls(0, 1)

    @classmethod
    def _param_schema(cls):
        return {
            "mean": {"type": "number"},
            "sd": {"type": "number"},
        }


@metadist(implements="core.truncated_normal", var_type="continuous")
class TruncatedNormalDistribution(ScipyDistribution):
    """Truncated normal distribution for floating point type.

    Parameters
    ----------
    lower:
        Lower bound of the truncated normal distribution.
    upper:
        Upper bound of the truncated normal distribution.
    mean:
        Mean of the non-truncated normal distribution.
    sd:
        Standard deviation of the non-truncated normal distribution.

    Examples
    --------
    >>> TruncatedNormalDistribution(lower=1.0, upper=3.5, mean=2.3, sd=5)
    """

    dist_class = truncnorm

    def __init__(self, lower: float, upper: float,
                 mean: float, sd: float):
        self.par = {"lower": lower, "upper": upper,
                    "mean": mean, "sd": sd}
        a, b = (lower-mean)/sd, (upper-mean)/sd
        self.dist = truncnorm(a=a, b=b, loc=mean, scale=max(sd, 1e-8))

    @classmethod
    def _fit(cls, values):
        lower = values.min() - 1e-8
        upper = values.max() + 1e-8
        return cls._fit_with_bounds(values, lower, upper)

    @classmethod
    def _fit_with_bounds(cls, values, lower, upper):
        def minimizer(param):
            mean, sd = param
            a, b = (lower-mean)/sd, (upper-mean)/sd
            dist = truncnorm(a=a, b=b, loc=mean, scale=sd)
            return -np.sum(dist.logpdf(values))

        x_start = [(lower+upper)/2, (upper-lower)/4]
        mean, sd = minimize(minimizer, x_start,
                             bounds=[(None, None),
                                     ((upper-lower)/100, None)]).x
        return cls(lower, upper, mean, sd)

    @classmethod
    def default_distribution(cls):
        return cls(-1, 2, 0, 1)

    @classmethod
    def _param_schema(cls):
        return {
            "lower": {"type": "number"},
            "upper": {"type": "number"},
            "mean": {"type": "number"},
            "sd": {"type": "number"},
        }


@metadist(implements="core.exponential", var_type="continuous")
class ExponentialDistribution(ScipyDistribution):
    """Exponential distribution for floating point type.

    This class implements the exponential distribution with the rate as its
    single parameter.

    Parameters
    ----------
    rate:
        Rate of the exponential distribution. This is equal to 1/mean of the distribution.

    Examples
    --------
    >>> ExponentialDistribution(rate=2.4)
    """

    dist_class = expon

    def __init__(self, rate: float):
        self.par = {"rate": rate}
        self.dist = expon(loc=0, scale=1/max(rate, 1e-8))

    @classmethod
    def _fit(cls, values):
        values = values.filter(values > 0)
        if len(values) == 0:
            return cls.default_distribution()
        return cls(rate=1/expon.fit(values, floc=0)[1])

    @classmethod
    def default_distribution(cls):
        return cls(1.0)

    @classmethod
    def _param_schema(cls):
        return {
            "rate": {"type": "number"}
        }



@metadist(implements="core.constant", var_type="continuous")
class ConstantDistribution(BaseConstantDistribution):
    """Constant distribution for floating point type.

    This class implements the constant distribution, so that it draws always
    the same value.

    Parameters
    ----------
    value: float
        Value that will be returned when drawn.

    Examples
    --------
    >>> ConstantDistribution(2.45)
    """

    @classmethod
    def default_distribution(cls) -> BaseDistribution:
        return cls(0.0)

    @classmethod
    def _param_schema(cls):
        return {
            "value": {"type": "number"}
        }
